(*#####################################################################*
 *
 * Module client pour le service de messagerie
 *
 * Toute la gestion du client y est inclue. On importe, en plus,
 * le cryptage des messages.
 *
 *#####################################################################*)


(***********************************************************************
 * Importation des modules
 ***********************************************************************)

open Types ;;
open Cryptage ;;
open Dialogues ;;
open Conversion ;;
open Observeur_connexions ;;
open Messages ;;
open Espaces ;;
open Gdk.Cursor ;;

(***********************************************************************
 * Variables globales
 ***********************************************************************)

(* si l'on est en mode debug ou pas *)
let debug = false ;;

(* port du serveur de messagerie *)
let port_serv_mess = ref 31237 ;;

(* numero et nom du client *)
let numero = ref 0 ;;
let nom_client = ref "" ;;

(* la liste des clients connects au service, tris par espace de
 * discussion. *)
let liste_espaces_par_client = ref ([] : (identite_client * (espace_discussion list)) list) ;;

(* la liste des espaces disponibles sur le serveur *)
let liste_espaces_discussion = ref [] ;;

(* le gestionnaire d'espaces *)
let gestionnaire_espaces = new gerant_espaces ;;

(* informations rseau concernant le client et la connexion *)
let in_channel_client = ref stdin ;;
let out_channel_client = ref stdout ;;


(* informations relatives au graphisme *)
let nom_fenetre = "OTchatche" ;;

let largeur_fenetre = 750 ;;
let hauteur_fenetre = 480 ;;

let taille_espacement = 20 ;;
let largeur_bordure = 30 ;;
let largeur_cadre_reception = 470 ;;
let largeur_espacement_table = 5 ;;
let bordure_livret = 0 ;;

let chemin_images =
   if Sys.os_type = "Unix" then
      (Sys.getenv "HOME") ^ "/.otchatche/"
   else
      ""
;;

let couleur_fond_fenetre = `NAME "steelblue4" ;;

let icone_espace_prive = "prive.xpm" ;;
let icone_espace_anonyme = "anonyme.xpm" ;;
let icone_espace_non_anonyme = "non_anonyme.xpm" ;;
let fichier_bordure_haute = "bordure_haute.xpm" ;;
let fichier_bordure_gauche = "bordure_gauche.xpm" ;;
let fichier_auteur = "auteur.xpm" ;;
let fichier_a_propos = "a_propos.xpm" ;;

(* dfinition du chemin des images dans le module Dialogues *)
definir_chemin_images_dialogues chemin_images ;;


(***********************************************************************
 * Partie graphique du client
 ***********************************************************************)

let inserer_element_liste = 
   let rec inserer = function e -> function position -> function index -> function
      | [] -> [e]
      | element :: reste ->
           if position = index then
              e :: element :: reste
           else
              element :: (inserer e position (index + 1) reste)
   in
   function e -> function position -> function liste ->
      inserer e position 0 liste
;;

let trier_liste = function liste ->
   List.sort
      (function e1 -> function e2 ->
          if e1 < e2 then
             -1
          else if e1 = e2 then
             0
          else
             1)
      liste
;;


let parser_chaine = 
   let rec supprimer_prefixe = function chaine ->
      if chaine = "" then
         ""
      else
         if chaine.[0] = ' ' then
            supprimer_prefixe (String.sub chaine 1 (String.length chaine - 1))
         else
            chaine

   and supprimer_suffixe = function chaine ->
      if chaine = "" then
         ""
      else
         if chaine.[String.length chaine - 1] = ' ' then
            supprimer_suffixe (String.sub chaine 0 (String.length chaine - 1))
         else
            chaine
               
   and verifier_absence_virgule = function chaine ->
      if chaine = "" then
         ""
      else
         if chaine.[0] = ',' then
            verifier_absence_virgule (String.sub chaine 1 (String.length chaine - 1))
         else
            (String.make 1 chaine.[0]) ^ 
            (verifier_absence_virgule (String.sub chaine 1 (String.length chaine - 1)))

   in
   function chaine ->
      verifier_absence_virgule (supprimer_suffixe (supprimer_prefixe chaine))
;;


let analyser_logins =
   let rec parseur = function chaine -> function login_courant ->
      if chaine = "" then
         []
      else
         if chaine.[0] = ',' then
            if parser_chaine login_courant <> "" then
               (parser_chaine login_courant) :: (parseur (String.sub chaine 1 (String.length chaine - 1)) "")
            else
               parseur (String.sub chaine 1 (String.length chaine - 1)) ""
         else
            parseur (String.sub chaine 1 (String.length chaine - 1)) (login_courant ^ (String.make 1 chaine.[0]))
   in
   function chaine ->
      parseur chaine ""
;;



let nom_espace = function
   | Espace_non_anonyme nom
   | Espace_anonyme nom
   | Espace_prive (nom, _) -> nom
;;


(* composant graphique pour grer l'affichage du tableau client/espace
 * o l'on peut voir quels clients sont dans quels espaces *)
class clients_espaces liste =
object(self)


   (* le conteneur global du composant *)
   val conteneur_pere = GBin.scrolled_window 
                           ~width:600 ~height:(min (22 * (List.length liste + 4)) 294)
                           ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ()

   (* le tableau des donnes pour l'tat des connexions *)
   val table_connexions = 
                      (* la bote contenant les tats *)
                      let creer_bouton_connexion = function () -> GButton.check_button ~active:false () 
                      in
                      (* la fonction permettant de changer l'tat d'une connexion *)
                      let changer_connexion = function bouton -> function connexion_active ->
                         bouton#set_active connexion_active ; bouton
                      in
                      (* permet de savoir si une connexion est active ou non *)
                      let connexion_est_active = function bouton ->
                         bouton#active
                      in
                      (* fournit le nom d'un client *)
                      let nom_de_client = function Identite (_, nom) -> nom
                      in
                      (* fournit le nom d'un espace *)
                      let nom_de_espace = function
                         | Espace_anonyme nom
                         | Espace_non_anonyme nom
                         | Espace_prive (nom, _) -> nom
                      in
                      (* fournit un ordre sur les clients *)
                      let inferieur_sur_clients = function c1 -> function c2 ->
                         let c1 = nom_de_client c1
                         and c2 = nom_de_client c2
                         in
                         if c1 < c2 then -1 else if c1 = c2 then 0 else 1
                      in
                      (* fourni un ordre sur les espaces *)
                      let inferieur_sur_espaces = function e1 -> function e2 ->
                         let e1 = nom_de_espace e1
                         and e2 = nom_de_espace e2
                         in
                         if e1 < e2 then -1 else if e1 = e2 then 0 else 1
                      in
                      new table_connexions
                         creer_bouton_connexion changer_connexion
                         connexion_est_active
                         nom_de_client nom_de_espace 
                         inferieur_sur_clients inferieur_sur_espaces

   (* les espaces slectionns *)
   val mutable liste_espaces_selectionnes = []

   (* permet de rcuprer le composant graphique *)
   method get_self = conteneur_pere#coerce

   (* permet de rcuprer la liste des espaces cochs *)
   method valider = function login ->
      if debug then (Printf.printf "%d espaces slectionns...\n" (List.length liste_espaces_selectionnes) ; flush stdout) ;
      (List.filter
          (function
              | Espace_non_anonyme _ | Espace_anonyme _ -> true
              | Espace_prive (_, liste) -> List.mem login liste)
          liste_espaces_selectionnes,
          liste_espaces_selectionnes)

   (* permet de remettre  jour l'tat des connexions *)
   method private mise_a_jour = function () ->
      let liste_clients = table_connexions#get_liste_clients ()
      and liste_espaces = table_connexions#get_liste_espaces ()
      and tableau = table_connexions#get_tableau ()
      in
      let largeur = List.length liste_espaces
      and hauteur = List.length liste_clients
      in
      let table = GPack.table 
                     ~rows:(hauteur + 3) ~columns:(largeur + 1) 
                     (*~row_spacings:0 ~col_spacings:0 *) ~homogeneous:false
                     ()
                     (* 2 lignes en plus pour les cases  cocher : slection des espaces *)
      in
      (* mise  jour des nom des espaces *)
      let label = GMisc.label ~text:"clients" ~justify:`CENTER ~width:50 () in
      let bouton_titre = GButton.button () in
      bouton_titre#set_relief `HALF ;
      bouton_titre#misc#set_sensitive false ;
      bouton_titre#add label#coerce ;
      table#attach ~left:0 ~top:0 bouton_titre#coerce ;

      for colonne = 0 to largeur - 1 do
         let espace_courant = List.nth liste_espaces colonne
         in
         let pack = GPack.hbox () in
         let pixmap_prive = GMisc.pixmap
                               (GDraw.pixmap_from_xpm
                                   ~file:(chemin_images ^ (
                                             match espace_courant with
                                                | Espace_prive _ -> icone_espace_prive
                                                | Espace_non_anonyme _ -> icone_espace_non_anonyme
                                                | Espace_anonyme _ -> icone_espace_anonyme))
                                   ())
                                ~packing:pack#add ()
         in
         let label = GMisc.label 
                        ~text:(nom_espace espace_courant)
                        ~justify:`CENTER ~packing:pack#add ()
         in
         let bouton_titre = GButton.button (*~width:70*) () 
         in
         bouton_titre#set_relief `HALF ;
         bouton_titre#misc#set_sensitive false ;
         bouton_titre#add pack#coerce ;
         table#attach ~left:(colonne + 1) ~top:0 bouton_titre#coerce
      done ;
      (* mise  jour des noms des clients *)
      for ligne = 0 to hauteur - 1 do
         let Identite (_, nom) = List.nth liste_clients ligne in
         let label = GMisc.label ~text:nom ~justify:`CENTER ()
         in
         let bouton_titre = GButton.button () 
         in
         bouton_titre#set_relief `HALF ;
         bouton_titre#misc#set_sensitive false ;
         bouton_titre#add label#coerce ;
         table#attach ~left:0 ~top:(ligne + 1) bouton_titre#coerce
      done ;
      let h = Array.length tableau in
      let l = if h = 0 then 0 else Array.length tableau.(0)
      in
      if debug then (Printf.printf "dbut de la construction du coeur de la table (%d, %d)...\n" h l ; flush stdout) ;
      (* mise  jour du coeur de la table *)
      for ligne = 0 to hauteur - 1 do
         for colonne = 0 to largeur - 1 do
            let rec chaine_n_espaces = function 
               | 0 -> ""
               | n -> " " ^ (chaine_n_espaces (n - 1))
            in
            let rec completer_nom = function nom -> function largeur ->
               if String.length nom >= largeur then
                  String.sub nom 0 largeur
               else
                  let manque = largeur - String.length nom
                  in
                  if manque mod 2 = 0 then
                     (chaine_n_espaces (manque / 2))
                     ^ nom ^ (chaine_n_espaces (manque / 2))
                  else
                     (chaine_n_espaces (manque / 2))
                     ^ nom ^ (chaine_n_espaces (manque / 2 + 1))
            in

            let taille_nom_espace_courant = String.length (nom_espace (List.nth liste_espaces colonne)) + 5 in
            let cellule = GList.clist ~columns:1 ~titles_show:false ()
            in
            ignore (cellule#append [""]) ;
            if (tableau.(ligne).(colonne))#active then
               cellule#set_cell ~text:(completer_nom "X" taille_nom_espace_courant) 0 0 ;
            cellule#set_shadow_type `ETCHED_IN ;
            cellule#misc#set_sensitive false ;
            table#attach ~left:(colonne + 1) ~top:(ligne + 1) cellule#coerce ;
         done
      done ;
      if debug then (Printf.printf "fin de la construction du coeur de la table...\n" ; flush stdout) ;
      (* mise  jour des espaces  slectionner *)
      for colonne = 0 to largeur - 1 do
         let espace_courant = List.nth liste_espaces colonne
         in
         let label_vide = GMisc.label ~text:"" ()
         and espace_est_selectionne = List.mem espace_courant liste_espaces_selectionnes
         in
         let pack =  GPack.hbox () in
         let bloc_vide = GMisc.label ~text:"" ~packing:pack#add () in
         let case_a_cocher = GButton.check_button ~active:espace_est_selectionne ~packing:pack#add ()
         in
         let reflexe = function () ->
            if case_a_cocher#active then
               liste_espaces_selectionnes <- espace_courant :: liste_espaces_selectionnes
            else
               liste_espaces_selectionnes <- List.filter (fun e -> e <> espace_courant) liste_espaces_selectionnes
         in
         ignore (case_a_cocher#connect#toggled ~callback:reflexe) ;
         table#attach ~left:(colonne + 1) ~top:(largeur + 1) label_vide#coerce ;
         table#attach ~left:(colonne + 1) ~top:(largeur + 2) pack#coerce
      done ;
      (* mise  jour du contenu de la scrolled_window *)
      let fils = conteneur_pere#children
      in
      if fils <> []  then
         conteneur_pere#remove (List.hd fils) ;
      conteneur_pere#add_with_viewport table#coerce


   (* permet d'ajouter un client dans le tableau *)
   method ajouter_client = function client ->
      table_connexions#ajouter_client client ;
      self#mise_a_jour ()

   (* permet de supprimer un client du tableau *)
   method supprimer_client = function client ->
      table_connexions#supprimer_client client ;
      self#mise_a_jour ()

   (* permet de supprimer un espace *)
   method supprimer_espace = function espace ->
      table_connexions#supprimer_espace espace ;
      self#mise_a_jour () ;

   (* permet d'ajouter un espace *)
   method ajouter_espace = function espace ->
      table_connexions#ajouter_espace espace ;
      self#mise_a_jour () ;
(*
   (* permet de dfinir les espaces auxquels sont connects les clients *)
   method definir_espaces_clients = function liste ->
      let liste = List.map
                     (fun (e, eg) ->
                         e, eg#clients_connectes)
                     liste
      in
      table_connexions#definir_tous_espaces liste ;
      self#mise_a_jour ()
 *)

   method definir_clients_espaces = function liste ->
      table_connexions#definir_tous_clients liste ;
      self#mise_a_jour ()

   (* permet de dfinir les espaces auxquels est connect le client *)
   method definir_espaces_client = function client -> function espaces ->
      table_connexions#definir_espaces_client client espaces ;
      self#mise_a_jour ()

   (* permet de grer la connexion d'un client  plusieurs espaces *)
   method connexion_espaces_client = function client -> function espaces ->
      List.iter (fun e -> table_connexions#definir_valeur client e true) espaces ;
      self#mise_a_jour ()

   (* permet de grer la dconnexion d'un client *)
   method deconnexion_espace_client = function client -> function espace ->
      table_connexions#definir_valeur client espace false ;
      self#mise_a_jour ()

   method private initialiser = function () ->
      for numero_client = 0 to List.length liste - 1 do
         let Identite (_, nom) as client_courant = fst (List.nth liste numero_client)
         in
         if debug then (Printf.printf "ajout du client (%s, %d)\n" nom numero_client ; flush stdout) ;
         self#ajouter_client client_courant
      done ;
      for numero_espace = 0 to List.length !liste_espaces_discussion - 1 do
         if debug then (Printf.printf "ajout de l'espace %d\n" numero_espace ; flush stdout) ;
         self#ajouter_espace (List.nth !liste_espaces_discussion numero_espace)
      done

   initializer
      if debug then (print_endline "Cration du tableau de visualisation..." ; flush stdout) ;
      self#initialiser () ;
      if debug then (print_endline "Fin de l'initialisation du tableau de visualisation..." ; flush stdout) ;
      self#definir_clients_espaces !liste_espaces_par_client ;
      self#mise_a_jour () 

end ;;


(***********************************************************************
 * Partie graphique du client
 ***********************************************************************)

let fenetre = GWindow.window
                 ~width:largeur_fenetre
                 ~height:hauteur_fenetre
                 ~position:`CENTER_ALWAYS
                 ~allow_shrink:false
                 ~allow_grow:false
                 ()
;;

let livret = GPack.notebook
                ~scrollable:true
                ~homogeneous_tabs:false
                ~show_border:false
                ()
;;

let champ_message = GEdit.entry ~width:largeur_cadre_reception () ;;

let bouton_envoyer = GButton.button ~label:titre_bouton_envoyer () ;;


let bordure_haute, bordure_gauche, auteur =
   let style_fenetre = fenetre#misc#style#copy
   in
   style_fenetre#set_bg [(`NORMAL, couleur_fond_fenetre)] ;
	fenetre#misc#set_style style_fenetre ;
   ((GMisc.pixmap
        (GDraw.pixmap_from_xpm
            ~file:(chemin_images ^ fichier_bordure_haute)
            ~window:fenetre ())
        ()),
    (GMisc.pixmap
        (GDraw.pixmap_from_xpm
            ~file:(chemin_images ^ fichier_bordure_gauche)
            ~window:fenetre ())
        ()),
    (GMisc.pixmap
        (GDraw.pixmap_from_xpm
            ~file:(chemin_images ^ fichier_auteur)
            ~window:fenetre ())
        ()))

in
let organisation1 = GPack.vbox ~packing:fenetre#add ()
in
organisation1#pack ~expand:false ~fill:false bordure_haute#coerce ;

let organisation2 = GPack.hbox ~packing:organisation1#add ()
in
let organisation_bordure = GPack.vbox ~packing:organisation2#pack ()
in
organisation_bordure#pack ~expand:false ~fill:false bordure_gauche#coerce ;
let pack_auteur = GPack.layout ~packing:organisation_bordure#add ()
in
pack_auteur#put auteur#coerce 0 0 ;
pack_auteur#event#add [`ENTER_NOTIFY; `LEAVE_NOTIFY; `BUTTON_PRESS] ;
ignore (pack_auteur#event#connect#enter_notify 
           ~callback:(fun _ ->
                         let type_curseur = `HAND2 in
                         let curseur = Gdk.Cursor.create type_curseur
                         in
                         Gdk.Window.set_cursor pack_auteur#misc#window curseur ;
                         true)) ;
ignore (pack_auteur#event#connect#leave_notify 
           ~callback:(fun _ ->
                         let type_curseur = `XTERM in
                         let curseur = Gdk.Cursor.create type_curseur
                         in
                         Gdk.Window.set_cursor pack_auteur#misc#window curseur ;
                         true)) ;

ignore (pack_auteur#event#connect#button_press 
           ~callback:(fun _ ->
                         let type_curseur = `HAND2 in
                         let curseur = Gdk.Cursor.create type_curseur
                         in
                         Gdk.Window.set_cursor pack_auteur#misc#window curseur ;
                         (new dialogue
                             ~titre_fenetre:"a propos..."
                             ~fichier_image:fichier_a_propos
                             ~modale:true
                             ~message:mess_a_propos
                             ~alignement_message:`CENTER
                             ~ligne_automatique:false
                             ~liste_boutons:[GButton.button ~label:"OK" ()]
                             ~liste_reflexes:[function () -> ()]
                             ~bouton_defaut:0
                             ~tuer:true)#show () ;
                         true)) ;

let organisation_fenetre = GPack.vbox
                              ~spacing:taille_espacement
                              ~border_width:largeur_bordure
                              ~packing:organisation2#add ()
in
organisation_fenetre#add livret#coerce ;

let organisation_message = GPack.hbox
                              ~spacing:taille_espacement
                              ~packing:organisation_fenetre#pack
                              ()
in
organisation_message#pack ~expand:false ~fill:false champ_message#coerce ;
organisation_message#add bouton_envoyer#coerce ;;


(***********************************************************************
 * Fonctions d'analyse lexicale
 ***********************************************************************)

(* dtermine si une chane de caractres est vide ou non. *)
let chaine_vide = function chaine ->
   chaine = ""
;;


(***********************************************************************
 * Fonctions basiques de communication avec le serveur
 ***********************************************************************)

(* permet de recevoir un vnement sans avoir  passer le canal
 * d'entre en argument. *)
let recevoir_evenement = function () ->
   lire_evenement !in_channel_client
;;

(* permet d'envoyer un vnement sans avoir  passer le canal
 * de sortie en argument. *)
let envoyer_evenement = function evenement ->
   ecrire_evenement !out_channel_client evenement
;;


(* crit un message en rcuprant l'espace courant et le contenu
 * du champ o l'on crit son message. *)
let ecrire_message = function () ->
   let message = parser_chaine (convertir_recuperation champ_message#text)
   in
   if not (chaine_vide message) then
      begin
         if debug then (Printf.printf
                           "J'envoie le message (%s) sur l'espace %s\n"
                           message
                           (match gestionnaire_espaces#espace_courant with
                               | Espace_non_anonyme nom
                               | Espace_anonyme nom
                               | Espace_prive (nom, _) ->
                                    nom) ;
                        flush stdout) ;

         envoyer_evenement
            (Envoi_message (
                Identite (!numero, !nom_client),
                gestionnaire_espaces#espace_courant,
                crypter message)) ;
         champ_message#set_text ""
      end
;;



(***********************************************************************
 * Fonctions de service
 ***********************************************************************)

(* permet d'imprimer une liste de type espace_discussion list sur la
 * sortie standard. Sert essentiellement au dbogage. *)
let rec imprimer_liste_espaces = function
   | [] -> flush stdout
   | Espace_anonyme nom :: reste ->
        print_endline (nom ^ extension_anonyme) ;
        imprimer_liste_espaces reste
   | Espace_non_anonyme nom :: reste ->
        print_endline nom ;
        imprimer_liste_espaces reste
   | Espace_prive (nom, _) :: reste ->
        print_endline (nom ^ extension_prive) ;
        imprimer_liste_espaces reste
;;


(* permet de sauvegarder une discussion dans un fichier. *)
let sauvegarder_discussion = function espace -> function nom_fichier ->
   (* on ouvre le fichier en lecture/ecriture pour l'utilisateur uniquement *)
   try
      let fichier = open_out_gen [Open_append; Open_text; Open_creat; Open_wronly] 384 nom_fichier
      in
      Printf.fprintf fichier "%s" (gestionnaire_espaces#discussion espace) ;
      close_out fichier
   with
      | Sys_error erreur ->
           let message = Printf.sprintf "%s\n\n(%s)"
                            mess_erreur_ecriture
                            erreur
           in
           (new dialogue_erreur message)#show ()

;;


(* filter les espaces qui ne nous sont pas accessibles. *)
let filtrer_espaces =
   let rec filtrer = function
      | [] -> []
      | (Espace_non_anonyme _ as espace) :: reste
      | (Espace_anonyme _ as espace) :: reste ->
           espace :: (filtrer reste)
      | (Espace_prive (_, liste_noms) as espace) :: reste ->
           if List.mem !nom_client liste_noms then
              espace :: (filtrer reste)
           else
              filtrer reste
   in
   function liste_espaces ->
      filtrer liste_espaces
;;

(* calcule l'intersection de deux listes *)
let rec intersection_listes = function l1 -> function
   | [] -> []
   | element :: reste ->
        if List.mem element l1 then
           element :: (intersection_listes l1 reste)
        else
           intersection_listes l1 reste
;;



(***********************************************************************
 * Rflexes attribus  la fentre principale
 ***********************************************************************)

let fermer_application = function () ->
   let nombre_espaces = gestionnaire_espaces#nombre_espaces_connectes in
   let liste_cases_cochees = ref []
   in
   let widget = GPack.table
                   ~rows:nombre_espaces
                   ~columns:3
                   ~row_spacings:largeur_espacement_table
                   ~col_spacings:largeur_espacement_table
                   ()
   in

   let rec reflexe_bouton = function case -> function champ -> function bouton ->
                            function espace -> function champ -> function decocher -> function () ->

      let reflexe_annuler = function fenetre -> function () ->
         if decocher then
            begin
               case#set_active false ;
               liste_cases_cochees := List.remove_assoc espace !liste_cases_cochees ;
               champ#misc#set_sensitive false ;
               bouton#misc#set_sensitive false
            end ;
         fenetre#destroy ()
      in

      let reflexe_ok = function fenetre -> function () ->
         let nom_fichier = parser_chaine (fenetre#get_filename)
         in
         if not (chaine_vide nom_fichier) then
            begin
               champ#set_text fenetre#get_filename ;
               liste_cases_cochees := List.remove_assoc espace !liste_cases_cochees ;
               liste_cases_cochees := (espace, fenetre#get_filename) :: !liste_cases_cochees ;
               fenetre#destroy ()
            end
         else
            reflexe_annuler fenetre ()
      in
      let entete = entete_fenetre_sauver ^
                   (match espace with
                       | Espace_anonyme nom
                       | Espace_non_anonyme nom
                       | Espace_prive (nom, _) ->
                            nom)
      in
      (new selection_fichiers entete reflexe_ok reflexe_annuler "*.txt")#show ()
   in

   let reflexe_case_espace = function case -> function espace -> function champ ->
                             function bouton -> function () ->
         if not case#active then
            begin
               champ#misc#set_sensitive false ;
               bouton#misc#set_sensitive false ;
               liste_cases_cochees := List.remove_assoc espace !liste_cases_cochees
            end
         else
            begin
               champ#misc#set_sensitive true ;
               bouton#misc#set_sensitive true ;
               reflexe_bouton case champ bouton espace champ true ()
            end ;
   in

   for i = 0 to nombre_espaces - 1 do
      let espace =
         try
            gestionnaire_espaces#espace_numero i
         with
            | Failure _ ->
                 failwith "fermer_application"
      in
      let case_espace = GButton.check_button
					            ~label:(convertir_affichage
                                      (match espace with
                                          | Espace_non_anonyme nom ->
                                               nom
                                          | Espace_anonyme nom ->
                                               nom ^ extension_anonyme
                                          | Espace_prive (nom, _) ->
                                               nom ^ extension_prive))
					            ()
	   in
      let champ = GEdit.entry ~editable:false ()
      in
      let bouton_selection = GButton.button ~label:"..." ()
      in
      ignore (case_espace#connect#toggled
                 ~callback:(reflexe_case_espace case_espace espace champ bouton_selection)) ;
      ignore (bouton_selection#connect#clicked
                 ~callback:(reflexe_bouton case_espace champ bouton_selection espace champ false)) ;

      champ#misc#set_sensitive false ;
      bouton_selection#misc#set_sensitive false ;

      widget#attach ~left:0 ~top:i case_espace#coerce ;
      widget#attach ~left:1 ~top:i champ#coerce ;
      widget#attach ~left:2 ~top:i bouton_selection#coerce ;

   done ;

   let reflexe_valider =
      let rec erreurs = ref false
      in
      let rec verifier = function
         | [] -> ()
         | (_, nom_fichier) :: reste ->
              try
                 let fichier = open_out_gen [Open_append; Open_text; Open_creat; Open_wronly] 384 nom_fichier
                 in
                 close_out fichier ;
                 verifier reste
              with
                 | Sys_error erreur ->
                      let message = Printf.sprintf "%s\n\n(%s)"
                                       mess_erreur_ecriture
                                       erreur
                      in
                      (new dialogue_erreur message)#show () ;
                      erreurs := true ;
                      verifier reste

      in
      function () ->
         verifier !liste_cases_cochees ;
         let valide = not !erreurs
         in
         erreurs := false ;
         valide
   in

   let reflexe_ok = function () ->
      let rec sauvegarder = function
         | [] -> ()
         | (espace, nom_fichier) :: reste ->
              sauvegarder_discussion espace nom_fichier ;
              sauvegarder reste
      in
      sauvegarder !liste_cases_cochees ;
      GMain.Main.quit () ;
      exit 0
   in

   champ_message#set_editable false ;
   bouton_envoyer#misc#set_sensitive false ;
   livret#misc#set_sensitive false ;
   if debug then (print_endline "Je vais lancer la bote de dialogues" ; flush stdout) ;
   let dialogue = new dialogue_question_depart
                     mess_cocher_esp_sauv
                     widget#coerce
                     reflexe_valider
                     reflexe_ok
   in
   dialogue#bouton_annuler#misc#hide () ;
   ignore (dialogue#fenetre#connect#destroy ~callback:(fun _ -> GMain.Main.quit () ; exit 0)) ;
   dialogue#show ()

;;



(***********************************************************************
 * Fonctions d'initialisation
 ***********************************************************************)

(* contruit les feuilles du livret en fonction des espaces auxquels on
 * se connecte. Si l'espace est un espace anonyme, la colonne des noms
 * n'est pas affiche. *)
let rec construire = function
   | [] -> ()
   | espace :: reste ->
        let position_insertion = gestionnaire_espaces#position_espace espace
        in

        (* pour la dnomination de l'espace : une image reprsentant
         * le caractre priv, anonyme ou non_anonyme de l'espace,
         * et son nom. *)
        let organisation_label = GPack.hbox ()
        in
        let dessin = GMisc.pixmap
                        (GDraw.pixmap_from_xpm
                            ~file:(chemin_images ^ (
                                      match espace with
                                         | Espace_prive _ -> icone_espace_prive
                                         | Espace_non_anonyme _ -> icone_espace_non_anonyme
                                         | Espace_anonyme _ -> icone_espace_anonyme))
                            ~window:fenetre ())
                        ~packing:organisation_label#add ()
        in
        let label = GMisc.label ~text:(match espace with
                                          | Espace_non_anonyme nom
                                          | Espace_anonyme nom
                                          | Espace_prive (nom, _) -> convertir_affichage nom)
                       ~packing:organisation_label#add ()
        in

        (* partie principale de l'espace. Si l'on a un espace anonyme,
         * il ne faut pas afficher le cadre des noms. *)
        let organisation_service = GPack.paned
                                      `HORIZONTAL
                                      ~border_width:bordure_livret
                                      (*~handle_size:10*) ()
        in
        let cadre_defilant_reception = GBin.scrolled_window
                                          ~width:largeur_cadre_reception
                                          ~hpolicy:`NEVER ~vpolicy:`AUTOMATIC
                                          (*~packing:organisation_service#add1*) ()
        in
        let panneau_reception = gestionnaire_espaces#panneau_messages espace
        in
        cadre_defilant_reception#add panneau_reception ;
        begin
           match espace with
              | Espace_anonyme _ -> ()
              | _ -> organisation_service#add1 cadre_defilant_reception#coerce
        end ;

        let cadre_defilant_noms = GBin.scrolled_window
                                     ~hpolicy:`NEVER ~vpolicy:`AUTOMATIC
                                     ~packing:organisation_service#add2 ()
        in
        let panneau_noms = gestionnaire_espaces#panneau_noms espace
        in
        cadre_defilant_noms#add panneau_noms ;

        livret#insert_page
           ~tab_label:organisation_label#coerce
           ~pos:position_insertion
           (match espace with
               | Espace_anonyme _ ->
                    cadre_defilant_reception#coerce
               | _ ->
                    organisation_service#coerce) ;

        gestionnaire_espaces#attribuer_reflexe espace
           (function bouton_souris ->
               if GdkEvent.Button.button bouton_souris = 3 then
                  begin
                     let popup_menu = GMenu.menu () in
                     let usine = new GMenu.factory popup_menu
                     in
                     let item_sauvegarder = GMenu.menu_item
                                               ~label:mess_menu_sauver
                                               ~packing:popup_menu#add ()
                     in
                     let item_effacer = GMenu.menu_item
                                           ~label:mess_menu_effacer
                                           ~packing:popup_menu#add ()
                     in
                     ignore (usine#add_separator ()) ;
                     let item_deconnecter = GMenu.menu_item
                                               ~label:mess_menu_deconnecter
                                               ~packing:popup_menu#add ()
                     in
                     let item_connecter = GMenu.menu_item
                                             ~label:mess_menu_connecter
                                             ~packing:popup_menu#add ()
                     in
                     ignore (usine#add_separator ()) ;
                     let item_creer_espace = GMenu.menu_item
                                                ~label:mess_menu_creer
                                                ~packing:popup_menu#add ()
                     in
                     let rec reflexe_sauvegarder = function () ->
                        let sauver = function fenetre -> function () ->
                           let nom_fichier = fenetre#get_filename
                           in
                           sauvegarder_discussion espace nom_fichier ;
                           fenetre#destroy ()
                        in
                        if gestionnaire_espaces#nombre_lignes_discussion espace = 0 then
                           (new dialogue_erreur
                               mess_erreur_sauv)#show ()
                        else
                           (new selection_fichiers
                               (mess_menu_sauver ^ "...")
                               sauver (fun _ -> fun () -> ())
                               "*.txt")#show ()
                     in
                     let reflexe_effacer = function () ->
                        gestionnaire_espaces#effacer_discussion espace
                     in
                     let reflexe_creer = function () ->
                        let widget = GPack.vbox ~spacing:taille_espacement()
                        in
                        let organisation_logins = GPack.hbox ()
                        in
                        let table_type = GPack.table
                                            ~rows:6 ~columns:2
                                            ~packing:widget#add ()
                        in
                        let label_type = GMisc.label ~text:titre_label_type ()
                        and label_logins = GMisc.label ~text:titre_label_logins ~packing:organisation_logins#add ()
                        and label_nom_espace = GMisc.label ~text:titre_label_noms ()
                        and bouton_non_anonyme = GButton.radio_button
                                                    ~label:titre_non_anonyme
                                                    ~active:true
                                                    ()
                        in
                        let bouton_anonyme = GButton.radio_button
                                                ~label:titre_anonyme
                                                ~group:bouton_non_anonyme#group
                                                ()
                        and bouton_prive = GButton.radio_button
                                              ~label:titre_bouton_prive
                                              ~group:bouton_non_anonyme#group
                                              ()
                        in
                        let champ_logins = GEdit.entry ~packing:organisation_logins#add ()
                        and champ_nom = GEdit.entry ()
                        in
                        ignore (champ_logins#event#connect#key_press
                                   (fun evenement ->
                                       let code = GdkEvent.Key.keyval evenement
                                       in
                                       if code = GdkKeysyms._Return then
                                          begin
                                             champ_nom#misc#grab_focus () ;
                                             true
                                          end
                                       else
                                          true)) ;
                        
                        let reflexe_bouton = function sensitif -> function () ->
                           champ_logins#misc#set_sensitive sensitif
                        in
                        bouton_non_anonyme#misc#grab_focus () ;
                        ignore (bouton_non_anonyme#connect#clicked ~callback:(reflexe_bouton false)) ;
                        ignore (bouton_anonyme#connect#clicked ~callback:(reflexe_bouton false)) ;
                        ignore (bouton_prive#connect#clicked ~callback:(reflexe_bouton true)) ;
                        champ_logins#misc#set_sensitive false ;
                        table_type#attach ~left:0 ~top:0 ~bottom:3 label_type#coerce ;
                        table_type#attach ~left:1 ~top:0 bouton_non_anonyme#coerce ;
                        table_type#attach ~left:1 ~top:1 bouton_anonyme#coerce ;
                        table_type#attach ~left:1 ~top:2 bouton_prive#coerce ;
                        table_type#attach ~left:1 ~top:3 organisation_logins#coerce ;
                        table_type#attach ~left:0 ~top:4 (GMisc.label ())#coerce ;
                        table_type#attach ~left:0 ~top:5 label_nom_espace#coerce ;
                        table_type#attach ~left:1 ~top:5 champ_nom#coerce ;
                        
                        let reflexe_valider = function () ->
                           let nom_espace = parser_chaine (convertir_recuperation champ_nom#text)
                           and liste_logins = analyser_logins (convertir_recuperation champ_logins#text)
                           in
                           if bouton_prive#active then
                              (not (chaine_vide nom_espace)) && liste_logins <> []
                           else
                              not (chaine_vide nom_espace)
                        in
                        
                        let reflexe_ok = function () ->
                           let nom_espace =
                              let nom = parser_chaine (convertir_recuperation champ_nom#text)
                              in
                              nom
                           in
                           let espace =
                              if bouton_anonyme#active then
                                 Espace_anonyme nom_espace
                              else if bouton_non_anonyme#active then
                                 Espace_non_anonyme nom_espace
                              else
                                 let liste_logins = analyser_logins (convertir_recuperation champ_logins#text)
                                 in
                                 let liste_logins =
                                    if List.mem !nom_client liste_logins then
                                       liste_logins
                                    else
                                       !nom_client :: liste_logins
                                 in
                                 Espace_prive (nom_espace, liste_logins)
                                    
                           in
                           ecrire_evenement !out_channel_client
                              (Demande_nouvel_espace espace)
                        in
                        let dialogue_demande = new dialogue_question
                                                  mess_definir_esp
                                                  widget#coerce
                                                  reflexe_valider
                                                  reflexe_ok
                                                  (fun _ -> ())
                        in
                        ignore (champ_nom#event#connect#key_press
                                   (fun evenement ->
                                       let code = GdkEvent.Key.keyval evenement
                                       in
                                       if code = GdkKeysyms._Return then
                                          begin
                                             if reflexe_valider () then
                                                begin
                                                   dialogue_demande#fenetre#destroy () ;
                                                   reflexe_ok ()
                                                end ;
                                             true
                                          end
                                       else
                                          true)) ;
                        dialogue_demande#show ()
                           
                     in
                     let reflexe_deconnecter = function () ->
                        if gestionnaire_espaces#nombre_espaces_connectes > 1 then
                           let ancienne_page = livret#current_page
                           in
                           if debug then (Printf.printf "********* dbut du rflexe deconnecter *************\n\n" ;
                           flush stdout) ;
                           
                           gestionnaire_espaces#supprimer_espace gestionnaire_espaces#espace_courant ;
                           ecrire_evenement
                              !out_channel_client
                              (Envoi_liste_espaces gestionnaire_espaces#liste_espaces_connectes) ;
                           ecrire_evenement
                              !out_channel_client
                              (Deconnexion_de (Identite (!numero, !nom_client), gestionnaire_espaces#espace_courant)) ;
                           livret#remove_page ancienne_page ;
                           if debug then 
                              begin
                                 Printf.printf "Liste des espaces connects :\n" ;
                                 imprimer_liste_espaces gestionnaire_espaces#liste_espaces_connectes ;
                                 Printf.printf "\nListe des espaces disponibles :\n" ;
                                 imprimer_liste_espaces !liste_espaces_discussion ;
                                 flush stdout ;
                           
                                 Printf.printf "********* fin du rflexe deconnecter *************\n\n" ;
                                 flush stdout
                              end ;
                        else
                           let message = mess_erreur_deconnecter
                           in
                           (new dialogue_erreur message)#show ()
                           
                     in
                     let reflexe_connecter = function () ->
                        if debug then
                           begin
                              Printf.printf "********* debut du rflexe connecter *************\n" ;
                              flush stdout ;
                              Printf.printf "\n\nListe des espaces connects :\n" ;
                              imprimer_liste_espaces gestionnaire_espaces#liste_espaces_connectes ;
                              Printf.printf "\n\nListe des espaces disponibles :\n" ;
                              imprimer_liste_espaces !liste_espaces_discussion ;
                              flush stdout
                           end ;
                        let liste_espaces_non_connectes =
                           trier_liste 
                              (filtrer_espaces
                                  (List.filter
                                      (fun e -> not (List.mem e gestionnaire_espaces#liste_espaces_connectes))
                                      !liste_espaces_discussion))
                        in
                        let liste_espaces_coches = ref []
                        in
                        
                        (* on dfinit le rflexe associ  chaque case  cocher de la liste *)
                        let reflexe_case = function numero_case -> function () ->
                           if List.mem numero_case !liste_espaces_coches then
                              (* si on est dans la liste, on doit s'en retirer car
                               * c'est qu'on tait slectionn et que l'on nous
                               * dslectionne *)
                              liste_espaces_coches :=
                              List.filter (fun x -> x <> numero_case) !liste_espaces_coches
                           else
                              (* on n'est pas dans la liste : on nous slectionne *)
                              liste_espaces_coches := numero_case :: !liste_espaces_coches ;
                           
                           liste_espaces_coches :=
                           List.sort
                              (fun n1 -> fun n2 -> if n1 > n2 then 1 else if n1 = n2 then 0 else -1)
                              !liste_espaces_coches
                        in
                        
                        let widget = GPack.vbox ()
                        in
                        
                        let construire_liste_espaces =
                           let rec construire = function index -> function
                              | [] -> ()
                              | espace :: reste ->
                                   let case_espace =
                                      match espace with
                                         | Espace_anonyme nom ->
                                              GButton.check_button
                                              ~label:(convertir_affichage (nom ^ extension_anonyme))
	                                           ~packing:widget#pack ()
                                         | Espace_non_anonyme nom ->
                                              GButton.check_button
                                              ~label:(convertir_affichage nom)
                                              ~packing:widget#pack ()
                                         | Espace_prive (nom, _) ->
                                              GButton.check_button
                                              ~label:(convertir_affichage (nom ^ extension_prive))
                                              ~packing:widget#pack ()
	                                in
	                                ignore (case_espace#connect#toggled ~callback:(reflexe_case index)) ;
                                   construire (index + 1) reste
                           in
                           function () ->
                              if liste_espaces_non_connectes = [] then
                                 failwith "erreur"
                              else
                                 construire 0 liste_espaces_non_connectes
                        in
                        
                        let reflexe_ok = function () ->
                           let liste = List.map
                                          (fun n -> List.nth liste_espaces_non_connectes n)
                                          !liste_espaces_coches
                           in
                           gestionnaire_espaces#ajouter_espaces liste ;
                           construire liste ;
                           ecrire_evenement
                              !out_channel_client
                              (Envoi_liste_espaces gestionnaire_espaces#liste_espaces_connectes) ;
                           ecrire_evenement !out_channel_client Demande_liste_clients ;
                           ecrire_evenement
                              !out_channel_client
                              (Connexion_a (Identite (!numero, !nom_client), liste))
                        in
                        try
                           construire_liste_espaces () ;
                           (new dialogue_question
                               mess_cocher_particip
                               widget#coerce
                               (fun _ -> true)
                               reflexe_ok
                               (fun _ -> ()))#show () ;
                           if debug then (Printf.printf "********* fin du rflexe connecter *************\n\n" ;
                           flush stdout)
                        with
                           | Failure "erreur" ->
                                let message = mess_erreur_connecter
                                in
                                (new dialogue_erreur message)#show ()
                     in
                     ignore (item_sauvegarder#connect#activate ~callback:reflexe_sauvegarder) ;
                     ignore (item_effacer#connect#activate ~callback:reflexe_effacer) ;
                     ignore (item_deconnecter#connect#activate ~callback:reflexe_deconnecter) ;
                     ignore (item_connecter#connect#activate ~callback:reflexe_connecter) ;
                     ignore (item_creer_espace#connect#activate ~callback:reflexe_creer) ;
                     popup_menu#popup ~button:1 ~time:0
                  end ;
               true) ;

        construire reste


let construire_feuillets = function () ->
   construire gestionnaire_espaces#liste_espaces_connectes
;;

(***********************************************************************
 * Gestion et traitement des vnements qui proviennent du serveur
 ***********************************************************************)

class gerant_evenements =
object (self)

   val mutable id_processus = Thread.self ()
   val mutable mort = false

   method demarrer = function () ->
      id_processus <- Thread.create self#recuperer_et_traiter () ;

   method terminer = function () ->
      if not mort then
         begin
            if debug then
               begin
                  print_endline "grant : je me tue." ;
                  Printf.printf "le numro du grant est : %d.\n" (Thread.id id_processus) ;
                  flush stdout
               end ;
            Thread.kill id_processus ;
            if debug then (Printf.printf "le numro du grant est : %d.\n" (Thread.id id_processus) ;
            flush stdout) ;
            mort <- true
         end
      else
         begin
            if debug then (print_endline "grant : je suis dj mort." ; flush stdout)
         end

   method private recuperer_evenement = function () ->
      if debug then (print_endline "grant : je suis en attente d'vnement..." ; flush stdout) ;
      let message = recevoir_evenement ()
      in
      if debug then (print_endline "grant : je viens de recevoir un vnement." ; flush stdout) ;
      message

   method private traiter_message = function evenement ->
      try
         match evenement with
            | Envoi_message (Identite (numero, nom), espace, message) ->
                 if debug then (print_endline "grant : je reois un message." ;
                 flush stdout) ;
                 gestionnaire_espaces#afficher_message espace nom (decrypter message)

            | Envoi_liste_clients liste_clients ->
                 if debug then (print_endline "grant : je reois un envoi de liste de clients." ;
                 flush stdout) ;
                 List.iter 
                    (fun (client, espaces) -> 
                        gestionnaire_espaces#ajouter_client client espaces)
                    liste_clients ;

            | Depart_client (Identite (numero, nom) as identite, liste_espaces) ->
                 if debug then (print_endline "grant : rception d'un dpart." ;
                 flush stdout) ;
                 gestionnaire_espaces#supprimer_client identite liste_espaces ;
                 List.iter 
                    (function
                        | Espace_anonyme _ ->
                             ()
                        | espace ->
                             gestionnaire_espaces#afficher_message 
                                espace 
                                nom_admin
                                (nom ^ mess_depart_client))
                    liste_espaces

            | Arrivee_client (Identite (numero, nom) as identite, liste_espaces) ->
                 if debug then (print_endline "grant : rception d'une arrive." ;
                 flush stdout) ;
                 gestionnaire_espaces#ajouter_client identite liste_espaces ;
                 List.iter 
                    (function
                        | Espace_anonyme _ ->
                             ()
                        | espace ->
                             gestionnaire_espaces#afficher_message
                                espace
                                nom_admin
                                (nom ^ mess_arrivee_client))
                    liste_espaces

            | Envoi_numero id ->
                 if debug then (print_endline "grant : je reois un envoi de numro." ;
                 flush stdout) ;
                 numero := id

            | Changement_espace (Identite (_, nom), ancien_espace, nouvel_espace) ->
                 if debug then (print_endline "grant : je reois un changement d'espace." ;
                 flush stdout) ;
                 if List.mem ancien_espace gestionnaire_espaces#liste_espaces_connectes then
                    begin
                       match ancien_espace with
                          | Espace_anonyme _ ->
                               ()
                          | _ ->
                               gestionnaire_espaces#afficher_message
                                  ancien_espace
                                  nom_admin
                                  (nom ^ mess_changt_espace)
                    end ;
                 if List.mem nouvel_espace gestionnaire_espaces#liste_espaces_connectes then
                    begin
                       match nouvel_espace with
                          | Espace_anonyme _ ->
                               ()
                          | _ ->
                               gestionnaire_espaces#afficher_message
                                  nouvel_espace
                                  nom_admin
                                  (nom ^ mess_arrivee_espace)
                    end

            | Refus_creation_espace espace ->
                 if debug then (print_endline "grant : je reois un refus de cration d'espace." ;
                 flush stdout) ;
                 let message = Printf.sprintf "%s %s %s"
                                  mess_erreur_refus_1
                                  (match espace with
                                      | Espace_non_anonyme nom ->
                                           nom
                                      | Espace_anonyme nom ->
                                           nom ^ extension_anonyme
                                      | Espace_prive (nom, _) ->
                                           nom ^ extension_prive)
                                  mess_erreur_refus_2
                 in
                 let message = convertir_affichage message
                 in
                 (new dialogue_information message)#show ()

            | Acceptation_creation_espace espace ->
                 if debug then (print_endline "grant : je reois une acceptation de cration d'espace." ;
                 flush stdout) ;

                 gestionnaire_espaces#ajouter_espace espace ;
                 construire [espace] ;
                 ecrire_evenement
                    !out_channel_client
                    (Envoi_liste_espaces gestionnaire_espaces#liste_espaces_connectes) ;
                 ecrire_evenement !out_channel_client Demande_liste_clients ;
                 ecrire_evenement
                    !out_channel_client
                    (Connexion_a (Identite (!numero, !nom_client), [espace]))

            | Notification_nouvel_espace espace ->
                 if debug then (print_endline "grant : je reois une notification d'une cration d'espace." ;
                 flush stdout) ;

                 begin
                    match espace with
                       | Espace_anonyme _
                       | Espace_non_anonyme _ ->
                            liste_espaces_discussion := espace :: !liste_espaces_discussion
                       | Espace_prive (_, liste) ->
                            if List.mem !nom_client liste then
                               liste_espaces_discussion := espace :: !liste_espaces_discussion
                 end

            | Notification_suppression_espace espace ->
                 if debug then (print_endline "grant : je reois une notification de suppression d'espace." ;
                 flush stdout) ;

                 liste_espaces_discussion :=
                    List.filter (fun e -> e <> espace) !liste_espaces_discussion ;

            | Connexion_a (Identite (_, nom) as id, liste) ->
                 if debug then (print_endline "grant : je reois une connexion  un espace." ;
                 flush stdout) ;

                 List.iter
                    (fun nouvel_espace ->
                        match nouvel_espace with
                           | Espace_anonyme _ ->
                                ()
                           | _ ->
                                gestionnaire_espaces#afficher_message
                                   nouvel_espace
                                   nom_admin
                                   (nom ^ mess_connexion_espace))
                    liste ;
                 gestionnaire_espaces#ajouter_client id liste

            | Deconnexion_de (Identite (_, nom) as id, espace) ->
                 if debug then (print_endline "grant : je reois une dconnexion d'un espace." ;
                 flush stdout) ;

                 begin
                    match espace with
                       | Espace_anonyme _ ->
                            ()
                       | _ ->
                            gestionnaire_espaces#afficher_message
                               espace
                               nom_admin
                               (nom ^ mess_deconnexion_espace)
                 end ;
                 gestionnaire_espaces#supprimer_client id [espace]

            | _ ->
                 failwith "cas impossible 6"
         with
            | exn ->
                 if debug then (print_endline (Printexc.to_string exn) ; flush stdout)

   method private recuperer_et_traiter = function () ->
      try
         while true do
            let message = self#recuperer_evenement ()
            in
            self#traiter_message message
         done
      with
         | End_of_file
         | Unix.Unix_error _  ->
              gestionnaire_espaces#afficher_message_partout nom_admin mess_erreur_coupure_1 ;
              gestionnaire_espaces#afficher_message_partout nom_admin mess_erreur_coupure_2 ;
              Thread.delay 5. ;
              fermer_application () ;
              mort <- true
                 (* le processus sort de la boucle automatiquement et
                  * termine *)

end
;;

let gerant_evenements = new gerant_evenements ;;


(***********************************************************************
 * Rflexes attribus  la fentre principale (2)
 ***********************************************************************)

let quitter = function () ->
   ecrire_evenement
      !out_channel_client
      (Depart_client (
          Identite (!numero, !nom_client),
          gestionnaire_espaces#liste_espaces_connectes)) ;

   gerant_evenements#terminer () ;
   if debug then (print_endline "Le grant est mort ; on peut terminer le programme." ; flush stdout) ;
   fermer_application ()
;;


fenetre#event#connect#delete (fun _ -> quitter () ; true) ;;
fenetre#event#connect#destroy (fun _ -> quitter () ; true) ;;

ignore (fenetre#event#connect#key_press
           (fun evenement ->
               let code = GdkEvent.Key.keyval evenement
               and etat = GdkEvent.Key.state evenement
               in
               if etat = [`CONTROL] then
                  if code = GdkKeysyms._Page_Down then
                     begin
                        livret#next_page () ;
                        champ_message#misc#grab_focus ()
                     end
                  else if code = GdkKeysyms._Page_Up then
                     begin
                        livret#previous_page () ;
                        champ_message#misc#grab_focus ()
                     end ;
               true)) ;;


(***********************************************************************
 * Fonctions principales
 ***********************************************************************)

(* lance la demande des espaces auxquels on peut se connecter, puis
 * poursuit en lanant la fentre finale. *)
let rec demandes = function in_channel -> function out_channel -> function liste ->

   let id_processus = ref (Thread.self ())
   and tableau_clients_espaces = new clients_espaces liste
   and champ_surnom = GEdit.entry ~max_length:15 ()
   in

   let widget = GPack.vbox ~spacing:30 () in
   let organisation_surnom = GPack.hbox ~packing:widget#add () in
   let label_surnom = GMisc.label ~text:titre_label_surnom ~packing:organisation_surnom#add ()
   in
   organisation_surnom#add champ_surnom#coerce ;
   widget#add tableau_clients_espaces#get_self ;

   let reflexe_validation = function () ->
      let surnom = parser_chaine (convertir_recuperation champ_surnom#text) in
      let resultat_espaces = tableau_clients_espaces#valider surnom in
      let liste_espaces_accessibles = fst resultat_espaces
      and liste_espaces_selectionnes = snd resultat_espaces
      in
      if surnom = nom_admin then
         begin
            (new dialogue_erreur mess_erreur_login)#show () ;
            false
         end
      else if surnom = "" then
         begin
            if debug then (print_endline "Le client n'a pas choisi de nom..." ;
            flush stdout) ;
            false
         end
      else if List.mem surnom 
         (List.map (fun (Identite (_, nom)) -> nom ) (fst (List.split !liste_espaces_par_client))) then
            begin
               (new dialogue_erreur mess_erreur_nom_deja_pris)#show () ;
               false
            end
      else if liste_espaces_selectionnes = [] then
         begin
            if debug then (print_endline "Le client n'a pas slectionn d'espaces..." ;
            flush stdout) ;
            false
         end 
      else if liste_espaces_accessibles <> liste_espaces_selectionnes then
         let liste_espaces_inaccessibles = List.filter 
                                             (function espace -> not (List.mem espace liste_espaces_accessibles))
                                             liste_espaces_selectionnes
         in
         let chaine_espaces_inaccessibles = chainer_liste 
                                               (List.map
                                                   (function espace -> nom_espace espace)
                                                   liste_espaces_inaccessibles)
         in
         let message = mess_erreur_selection_1 ^ " \"" ^ surnom ^ "\" " ^ 
                          (if List.length liste_espaces_inaccessibles = 1 then
                              mess_erreur_selection_2 
                           else
                              mess_erreur_selection_3) ^ " " ^
                           chaine_espaces_inaccessibles ^ "."
         in
         (new dialogue_erreur message)#show () ;
         false
      else
         true
   in

   let reflexe_apres_validation = function () ->
      gestionnaire_espaces#ajouter_espaces (snd (tableau_clients_espaces#valider "")) ;
      Thread.kill !id_processus ;
      nom_client := parser_chaine (convertir_recuperation champ_surnom#text) ;
      if debug then (imprimer_liste_espaces gestionnaire_espaces#liste_espaces_connectes) ;
      construire_feuillets () ;
      gestionnaire_espaces#definir_espace_courant
         begin
            try
               List.nth gestionnaire_espaces#liste_espaces_connectes livret#current_page
            with
               | Failure _ ->
                    failwith "demandes"
         end ;
      ecrire_evenement out_channel (Envoi_nom !nom_client) ;
      ecrire_evenement out_channel (Envoi_liste_espaces gestionnaire_espaces#liste_espaces_connectes) ;
      liste_espaces_par_client := [] ;
      ecrire_evenement out_channel Demande_liste_clients ;
      gerant_evenements#demarrer () ;
      ignore (livret#connect#switch_page ~callback:
                 (let ancienne_page = ref 0
                  in
                  function nouvelle_page ->
                     if debug then (Printf.printf "********* dbut du rflexe switch *************\n" ;
                     flush stdout) ;
                     let feuille_courante = livret#current_page in
                     let nouvelle_page =
                        (* correction de bug lors de la suppression de page *)
                        if feuille_courante = -1 && !ancienne_page = 0 then
                           0
                        else
                           nouvelle_page
                     in
                     let nouvel_espace =
                        try
                           List.nth gestionnaire_espaces#liste_espaces_connectes nouvelle_page
                        with
                           | Failure _ ->
                                if debug then
                                   (Printf.printf "\n\nfeuille_courante = %d\n" feuille_courante ;
                                    Printf.printf "nouvelle_page = %d\n\n" nouvelle_page) ;
                                failwith "rflexe switch_page livret"
                     in
                     if debug then
                        begin
                           Printf.printf "\n\nfeuille_courante = %d\n" feuille_courante ;
                           Printf.printf "nouvelle_page = %d\n\n" nouvelle_page ;
                           flush stdout ;
                           Printf.printf "\n\nje clique sur la page n%d\n" nouvelle_page ;
                           flush stdout ;
                           Printf.printf "le nom de l'ancien espace est \"%s\"\n\n"
                              (match gestionnaire_espaces#espace_courant with
                                  | Espace_non_anonyme nom
                                  | Espace_anonyme nom
                                  | Espace_prive (nom, _) ->
                                       nom) ;
                           Printf.printf "le nom du nouvel espace est \"%s\"\n\n"
                              (match nouvel_espace with
                                  | Espace_non_anonyme nom
                                  | Espace_anonyme nom
                                  | Espace_prive (nom, _) ->
                                       nom) ;
                           flush stdout
                        end ;
                     ecrire_evenement
                        !out_channel_client
                        (Changement_espace (Identite (!numero, !nom_client),
                                            gestionnaire_espaces#espace_courant,
                                            nouvel_espace)) ;
                     ancienne_page := nouvelle_page ;
                     gestionnaire_espaces#definir_espace_courant nouvel_espace ;
                     champ_message#misc#grab_focus () ;
                     if debug then (Printf.printf "********* fin du rflexe switch *************\n\n" ;
                     flush stdout))) ;

      (* mise en place des fonctionnalits des boutons *)
      ignore (bouton_envoyer#connect#clicked ecrire_message) ;
      ignore (champ_message#event#connect#key_press
                 (fun evenement ->
                     let code = GdkEvent.Key.keyval evenement
                     in
                     if code = GdkKeysyms._Return then
                        begin
                           ecrire_message () ;
                           true
                        end
                     else
                        false)) ;

      champ_message#misc#grab_focus () ;

      (* affichage de la fenetre *)
      fenetre#set_title (convertir_affichage (nom_fenetre ^ " (" ^ !nom_client ^ ")")) ;
      fenetre#set_position `CENTER_ALWAYS ;
      fenetre#show ()

   in
   let reflexe_fermeture = function () ->
      envoyer_evenement (Depart_client (Identite (!numero, !nom_client), !liste_espaces_discussion))
   in
   let dialogue = new dialogue_question_depart2
                     mess_cocher_particip
                     widget#coerce
                     reflexe_validation
                     reflexe_apres_validation
                     reflexe_fermeture
   in
   let processus_ecoute = function () ->
      try
         while true do
            if debug then (Printf.printf "Je suis en attente d'vnements de la part du serveur.\n" ;
            flush stdout) ;
            let evenement = lire_evenement in_channel
            in
            match evenement with
               | Notification_nouvel_espace espace ->
                    if debug then (Printf.printf "Je reois une cration d'espace.\n\n" ;
                    flush stdout) ;
                    tableau_clients_espaces#ajouter_espace espace
                       
               | Notification_suppression_espace espace ->
                    if debug then (Printf.printf "Je reois une suppression d'espace.\n\n" ;
                    flush stdout) ;
                    tableau_clients_espaces#supprimer_espace espace

               | Envoi_message _ ->
                    if debug then (Printf.printf "Je reois un envoi de message !\n" ;
                    flush stdout)

               | Envoi_liste_clients liste ->
                    if debug then (Printf.printf "Je reois un envoi de liste !\n" ;
                    flush stdout) ;
                    tableau_clients_espaces#definir_clients_espaces liste

               | Depart_client (client, _) ->
                    if debug then (Printf.printf "Je reois un dpart de client !\n" ;
                    flush stdout) ;
                    tableau_clients_espaces#supprimer_client client

               | Arrivee_client (client, espaces) ->
                    if debug then (Printf.printf "Je reois une arrive de client !\n" ;
                    flush stdout) ;
                    tableau_clients_espaces#ajouter_client client ;
                    tableau_clients_espaces#definir_espaces_client client espaces

               | Envoi_numero _ ->
                    if debug then (Printf.printf "Je reois un envoi de numro !\n" ;
                    flush stdout)

               | Changement_espace _ ->
                    if debug then (Printf.printf "Je reois un changement d'espace !\n" ;
                    flush stdout)

               | Refus_creation_espace _ ->
                    if debug then (Printf.printf "Je reois un refus de cration !\n" ;
                    flush stdout)

               | Acceptation_creation_espace _ ->
                    if debug then (Printf.printf "Je reois une acceptation de cration !\n" ;
                    flush stdout)

               | Connexion_a (client, liste) ->
                    if debug then (Printf.printf "Je reois une connexion  un espace !\n" ;
                    flush stdout) ;
                    tableau_clients_espaces#connexion_espaces_client client liste

               | Deconnexion_de (client, espace) ->
                    if debug then (Printf.printf "Je reois une dconnexion d'un espace !\n" ;
                    flush stdout) ;
                    tableau_clients_espaces#deconnexion_espace_client client espace

               | _ ->
                    if debug then (Printf.printf "Je reois un vnement non attendu !\n" ;
                    flush stdout)
         done
      with
         | End_of_file 
         | Unix.Unix_error _ ->
              let fermeture = function () ->
                 dialogue#fenetre#misc#hide () ;
                 lancer_demande ()
              in
              let dialogue_erreur =
                 new dialogue_erreur2 
                    (convertir_affichage mess_erreur_coupure_1) 
                    (fun _ -> ())
              in
              ignore (dialogue_erreur#fenetre#connect#destroy ~callback:fermeture) ;
              ignore (dialogue_erreur#fenetre#event#connect#delete ~callback:(fun _ -> false)) ;
              dialogue_erreur#show ()
   in


   id_processus := Thread.create processus_ecoute () ;
   (*ignore (dialogue#fenetre#connect#destroy ~callback:(fun _ -> GMain.Main.quit () ; exit 0)) ;*)
   champ_surnom#misc#grab_focus () ;
   dialogue#show ()



(* initialise la connexion avec le serveur (construction de la socket,
 * rcupration des canaux, ...) et rcupre la liste des espaces
 * actuellement disponibles sur le serveur. *)
and lancer_connexion = function nom_serveur ->

   try
      (* connexion du client au serveur de messagerie *)
      if debug then (print_endline "client : demande de connexion avec le serveur intermdiaire..." ;
      flush stdout) ;

      let addr_serv_mess = nom_serveur
      in

      if debug then (print_endline "client : je me connecte avec le serveur de messagerie." ;
      flush stdout) ;

      (* rcupration de l'adresse du serveur de messagerie *)
      let addr_serv_mess =
         try Unix.inet_addr_of_string addr_serv_mess with
            | Failure "inet_addr_of_string" ->
                 (* l'adresse tape n'est pas une adresse inet valide : c'est un nom *)
                 try (Unix.gethostbyname addr_serv_mess).Unix.h_addr_list.(0) with
                    | Not_found ->
                         if debug then (print_endline "client : le serveur de messagerie n'est pas connu." ;
                         flush stdout) ;
                         failwith "erreur"
                    | exn ->
                         if debug then
                            begin
                               print_endline "client : erreur avec l'adresse du serveur de messagerie." ;
                               print_endline ("            " ^ (Printexc.to_string exn)) ;
                               flush stdout
                            end ;
                         failwith "erreur"

      in
      let addr_sock_serv_mess = Unix.ADDR_INET (addr_serv_mess, !port_serv_mess) in

      (* renvoi des canaux de communication du serveur de messagerie *)
      let canaux =
         begin
            try Unix.open_connection addr_sock_serv_mess with
               | Unix.Unix_error _ as e ->
                    if debug then 
                       begin
                          print_endline ("client : le serveur de messagerie ne peut vous " ^
                                         "ouvrir une connexion.") ;
                          print_endline (Printexc.to_string e) ;
                          flush stdout
                       end ;
                    failwith "erreur"
               | exn ->
                    if debug then 
                       begin
                          print_endline "client : erreur avec l'ouverture de la connexion." ;
                          print_endline ("            " ^ (Printexc.to_string exn)) ;
                          flush stdout
                       end ;
                    failwith "erreur"
         end
      in
      in_channel_client := fst canaux ;
      out_channel_client := snd canaux ;

      if debug then (print_endline "client : les canaux viennent d'tre rcuprs." ;
      flush stdout) ;

      (* on rcupre les espaces disponibles pour les choisir *)
      let evenement_numero = lire_evenement !in_channel_client in
      let espaces_disponibles = lire_evenement !in_channel_client in
      let noms_connectes = lire_evenement !in_channel_client
      in
      match evenement_numero, espaces_disponibles, noms_connectes with
         | Envoi_numero n, 
           Envoi_liste_espaces espaces_disponibles,
           Envoi_liste_clients liste ->
              numero := n ;
              liste_espaces_discussion := espaces_disponibles ;
              liste_espaces_par_client := liste ;
              if debug then (print_endline "Donnes rcupres : on passe au panneau des demandes..." ; flush stdout) ;
              demandes !in_channel_client !out_channel_client liste
         | _ ->
              failwith "cas impossible 7"

   with
      | Unix.Unix_error _
      | Failure ("erreur") ->
           let message = mess_erreur_serv_abs
           in
           (new dialogue_attention
               message
               (fun _ ->
                   lancer_demande ())
               (fun _ ->
                   gerant_evenements#terminer () ;
                   GMain.Main.quit () ;
                   exit 0))#show ()



(* initialise la demande des informations. Demande le nom du serveur. *)
and lancer_demande = function () ->

   (* Demande d'informations relatives  la connexion *)
   let organisation = GPack.hbox () in
   let table = GPack.table ~rows:2 ~columns:2 ~packing:organisation#add ()
   in
   let label_serveur = GMisc.label ~text:titre_label_serveur ~packing:(table#attach ~left:0 ~top:0) ()
   and champ_serveur = GEdit.entry ~packing:(table#attach ~left:1 ~top:0) ()
   and label_port = GMisc.label ~text:titre_port_serveur ~packing:(table#attach ~left:0 ~top:1) ()
   and champ_port = GEdit.entry ~packing:(table#attach ~left:1 ~top:1) ()
   in
   let reflexe_validation = function _ ->
      try
         let numero_port = int_of_string champ_port#text
         in
         (not (chaine_vide (parser_chaine champ_serveur#text))) &&
         (0 <= numero_port && numero_port <= 32768)
      with
         | _ -> false
   in

   let reflexe_ok = function _ ->
      port_serv_mess := int_of_string champ_port#text ;
      lancer_connexion champ_serveur#text
   in
   let message = mess_dem_infos_connexion
   in
   let dialogue_demandes = new dialogue_question_depart
                              message
                              organisation#coerce
                              reflexe_validation
                              reflexe_ok
   in
   champ_serveur#misc#grab_focus () ;
   champ_port#set_text (string_of_int !port_serv_mess) ;
   ignore (champ_serveur#event#connect#key_press
              (fun evenement ->
                  let code = GdkEvent.Key.keyval evenement
                  in
                  if code = GdkKeysyms._Return then
                     begin
                        champ_port#misc#grab_focus () ;
                        true
                     end
                  else
                     true)) ;
   ignore (champ_port#event#connect#key_press
              (fun evenement ->
                  let code = GdkEvent.Key.keyval evenement
                  in
                  if code = GdkKeysyms._Return then
                     begin
                        if not (reflexe_validation ()) then
                           true
                        else
                           begin
                              dialogue_demandes#fenetre#misc#hide () ;
                              reflexe_ok () ;
                              true
                           end
                     end
                  else
                     true)) ;

   ignore (dialogue_demandes#fenetre#connect#destroy
              ~callback:(fun _ -> GMain.Main.quit () ; exit 0)) ;
   dialogue_demandes#show ()


;;


(***********************************************************************
 * Lancement du client
 ***********************************************************************)

fenetre#show () ;
(if debug then print_endline else fun _ -> ()) (GMain.Main.init ~setlocale:true ()) ;;
lancer_demande () ;
flush stdout ;;
GtkThread.main ()
