(*#####################################################################*
 *
 * Module Conversion
 *
 * Permet de convertir les chanes de caractres rcupres de
 * l'interface GTK pour Windows en chanes de caractres Linux,
 * et rciproquement.
 *
 * La chane Windows dsigne la chane qui est affiche ou qui doit
 * tre affiche sur la fentre GTK pour Windows.
 * La chane Linux dsigne la chane classique.
 *
 *#####################################################################*)


(**********************************************************************
 * Variables
 **********************************************************************)

(* donne les intervalles des plages de caractres *)
let borne_inf_plage_1 = 0x80 ;;
let borne_sup_plage_1 = 0xBF ;;
let borne_inf_plage_2 = 0xC0 ;;
let borne_sup_plage_2 = 0xFF ;;

(* donne les index des plages 1 et 2 *)
let caractere_plage_1 = 0xC2 ;;
let caractere_plage_2 = 0xC3 ;;


(**********************************************************************
 * Fonctions
 **********************************************************************)

(* convertit la chane windows en la chane linux qui lui correspond,
 * ie :
 *    o si le 1er caractre de la chane est caractere_plage_1, on
 *      garde le second sans modification ;
 *    o si le 1er caractre de la chaine est caractere_plage_2, on
 *      ajoute au second caractre le dcalage des plages
 *      (borne_inf_plage_2 - borne_inf_plage_1).
 *)
let chaine_windows_vers_linux =
   let convertir_char = function s ->
      let code_char_0 = Char.code s.[0]
      and char_1 = s.[1]
      in
      if code_char_0 = caractere_plage_2 then
         Printf.sprintf "%c"
            (char_of_int (Char.code char_1 + (borne_inf_plage_2 - borne_inf_plage_1)))
      else if code_char_0 = caractere_plage_1 then
         Printf.sprintf "%c" char_1
      else
         s
   in
   let rec convertir = function
      | "" -> ""
      | s ->
           if String.length s = 1 then
              s
           else
              let prefixe_1 = String.sub s 0 1
              and prefixe_2 = String.sub s 1 1
              and suffixe = String.sub s 2 (String.length s - 2) in
              let prefixe = prefixe_1 ^ prefixe_2 in
              let conv = convertir_char prefixe
              in
              if conv = prefixe then
                 prefixe_1 ^ (convertir (prefixe_2 ^ suffixe))
              else
                 conv ^ (convertir suffixe)
   in
   function chaine ->
      convertir chaine
;;


(* convertit la chane linux en la chane windows qui lui correspond,
 * ie :
 *    o si le caractre appartient  la plage n1, on lui rajoute
 *      caractere_plage_1 devant ;
 *    o si le caractre appartient  la plage n2, on lui rajoute
 *      caractere_plage_2 devant, et on lui ajoute le dcalage
 *      des plages (borne_inf_plage_2 - borne_inf_plage_1).
 *)
let chaine_linux_vers_windows =
   let convertir_char = function s ->
      let char = s.[0] in
      let code_char = Char.code char
      in
      if borne_inf_plage_1 <= code_char &&
         code_char <= borne_sup_plage_1 then
         Printf.sprintf "%c%c"
            (char_of_int caractere_plage_1)
            char
      else if borne_inf_plage_2 <= code_char &&
              code_char <= borne_sup_plage_2 then
         Printf.sprintf "%c%c"
            (char_of_int caractere_plage_2)
            (char_of_int (code_char - (borne_inf_plage_2 - borne_inf_plage_1)))
      else
         s

   in
   let rec convertir = function
      | "" -> ""
      | s ->
           if String.length s = 1 then
              convertir_char s
           else
              let prefixe = String.sub s 0 1
              and suffixe = String.sub s 1 (String.length s - 1) in
              let conv = convertir_char prefixe
              in
              conv ^ (convertir suffixe)
   in
   function chaine ->
      convertir chaine
;;

(* convertit une chane linux ou windows pour qu'elle soit affiche
 * correctement par l'interface. *)
let convertir_affichage = function s ->
   if Sys.os_type = "Unix" then
      s
   else
      chaine_linux_vers_windows s
;;

(* convertit une chane tape par l'utilisateur pour la transformer
 * par une chane linux. *)
let convertir_recuperation = function s ->
   if Sys.os_type = "Unix" then
      s
   else
      chaine_windows_vers_linux s
;;
