(*#####################################################################*
 *
 * Module client pour le service de messagerie
 *
 * Toute la gestion du client y est inclue. On importe, en plus,
 * le cryptage des messages.
 *
 *#####################################################################*)


open Types ;;
open Conversion ;;
open Messages ;;
open Dates ;;

(***********************************************************************
  * Les variables
 ***********************************************************************)

let couleur_defaut = `DEFAULT ;;
let couleur_texte_administrateur = `NAME "red" ;;
let couleur_impaire = couleur_defaut ;;
let couleur_paire = `NAME "grey90" ;;



(***********************************************************************
  * Fonctions utilitaires
 ***********************************************************************)

(* permet de trouver la premire ligne d'une clist contenant, en premire
 * cellule, le texte pass en argument.
 * Renvoie Not_found si le client n'a pas t trouv *)
let chercher_ligne = function texte -> function clist ->
   let rec chercher = function ligne ->
      if ligne = clist#rows then
         raise Not_found
      else
         if clist#cell_text ligne 0 = texte then
            ligne
         else
            chercher (ligne + 1)
   in
   chercher 0
;;


(* permet de couper une chane de caractres en plusieurs morceaux,
 * tant donn la longueur maximale d'un morceau et la marge de 
 * manoeuvre autour de la taille maximale *)
let morceler_chaine = function chaine -> function longueur -> function marge ->
   let rec couper = function chaine -> function endroit_coupure ->
      if String.length chaine <= endroit_coupure then
         [chaine]
      else if abs (endroit_coupure - longueur) > marge then
         (String.sub chaine 0 longueur)
         :: (couper 
                (String.sub chaine longueur (String.length chaine - longueur)) 
                longueur)
      else
         let caractere_avant_coupure = chaine.[endroit_coupure - 1]
         and caractere_apres_coupure = chaine.[endroit_coupure]
         in
         match caractere_avant_coupure, caractere_apres_coupure with
            | _ , ('?'|'!'|'.') ->
                 couper chaine (endroit_coupure + 1)
            | ('?'|'!'|'.'), ' ' ->
                 (String.sub chaine 0 endroit_coupure)
                 :: (couper 
                        (String.sub chaine (endroit_coupure + 1) (String.length chaine - endroit_coupure - 1)) 
                        longueur)                 
            | _, ' ' -> 
                 (String.sub chaine 0 endroit_coupure)
                 :: (couper 
                        (String.sub chaine (endroit_coupure + 1) (String.length chaine - endroit_coupure - 1)) 
                        longueur)
            | _, (','|';'|':') ->
                 (String.sub chaine 0 (endroit_coupure + 1))
                 :: (couper 
                        (String.sub chaine (endroit_coupure + 1) (String.length chaine - endroit_coupure - 1)) 
                        longueur)
            | _, _ -> 
                 couper chaine (endroit_coupure - 1)
   in
   couper chaine longueur
;;




(***********************************************************************
  * La classe espace
 ***********************************************************************)

class espace espace_discussion =
object (self)


   (* les attributs ***************************************)

   (* contient la liste des clients connects  l'espace courant *)
   val mutable liste_clients_connectes = ([] : identite_client list)

   (* le numro de l'espace (son ordre dans le feuillet) *)
   val mutable numero_page = -1

   (* dfinit la couleur du message courant *)
   val mutable couleur_message_courant = couleur_impaire

   (* le cadre contenant les messages *)
   val panneau_messages = GList.clist
                             ~columns:2
                             ~titles:[titre_liste_nom; titre_liste_message]
                             ~selection_mode:`SINGLE
                             ~titles_active:false
                             ()


   (* le cadre contenant le nom des clients connects *)
   val panneau_noms = GList.clist
                         ~columns:2
                         ~titles:[titre_liste_id; titre_liste_connectes]
                         ~selection_mode:`SINGLE
                         ~sort_column:1
                         ~auto_sort:true
                         ~titles_active:false
                         ()

   (* largeur du texte et marge de manoeuvre pour la dcoupe *)
   val largeur_texte = 
    match espace_discussion with
       | Espace_anonyme _ -> 100
       | _ -> 50

   val marge_manoeuvre = 6


   (* les mthodes ****************************************)

   (* permet de rcuprer le nom de l'espace *)
   method nom = 
      match espace_discussion with
         | Espace_non_anonyme nom
         | Espace_anonyme nom
         | Espace_prive (nom, _) -> nom

   (* permet de rcuprer le type de l'espace *)
   method espace = espace_discussion

   (* permet de rcuprer le numro de l'espace *)
   method numero = numero_page

   (* permet de rcuprer la liste des clients connects *)
   method clients_connectes = liste_clients_connectes

   (* permet de rcuprer le panneau des messages *)
   method panneau_messages = panneau_messages#coerce

   (* permet de rcuprer le panneau des noms *)
   method panneau_noms = panneau_noms#coerce

   (* permet de rcuprer le nombre de lignes *)
   method nombre_lignes = panneau_messages#rows


   (* permet d'effacer le panneau des messages *)
   method effacer_messages = function () ->
      panneau_messages#clear ()

   (* permet d'attribuer un rflexe au panneau des messages *)
   method attribuer_reflexe = function f ->
      ignore (panneau_messages#event#connect#button_press ~callback:f)


   (* permet d'ajouter un client dans la liste des clients connects *)
   method ajouter_client = function Identite (id, nom) as client ->
      if not (List.mem client liste_clients_connectes) then
         begin
            liste_clients_connectes <- client :: liste_clients_connectes ;
            ignore (panneau_noms#append [string_of_int id; convertir_affichage nom])
         end

   (* permet de supprimer un client de la liste des clients connects *)
   method supprimer_client = function Identite (id, nom) as client ->
      if List.mem client liste_clients_connectes then
         let ligne = chercher_ligne (string_of_int id) panneau_noms
         in
         liste_clients_connectes <- List.find_all (fun c -> c <> client) liste_clients_connectes ;
         panneau_noms#remove ligne

   (* permet d'afficher le message d'un client *)
   method afficher_message = function nom -> function chaine ->
      let rec afficher = function nom -> function couleur_texte -> function
         | [] -> ()
         | morceau :: reste ->
              let ligne = panneau_messages#append [nom; morceau]
              in
              panneau_messages#set_row 
                 ~foreground:couleur_texte 
                 ~background:couleur_message_courant 
                 ligne ;
              afficher "" couleur_texte reste
      in
      let nom_auteur = convertir_affichage nom
      and morceaux = List.map 
                        (fun c -> convertir_affichage c) 
                        (morceler_chaine chaine largeur_texte marge_manoeuvre)
      in
      afficher 
         nom_auteur 
         (if nom = nom_admin then couleur_texte_administrateur else couleur_defaut)
         morceaux ;
      couleur_message_courant <-
         if couleur_message_courant = couleur_impaire then
            couleur_paire
         else
            couleur_impaire

   (* permet de rcuprer la discussion sous forme d'une seule chaine de caractres *)
   method discussion = 
      let chaine = ref ""
      in
      let (minutes, heures, nom_jour, jour, mois, annee) = date_actuelle () in
      let date = Printf.sprintf "%s %d/%d/%d" nom_jour jour mois annee
      and heure = Printf.sprintf "%d:%s%d" heures (if minutes <= 9 then "0" else "") minutes
      in
      chaine := 
      Printf.sprintf
         "%s\n* discussion sur l'espace %s\n* termine le %s\n*  %s.\n%s\n\n"
         "************************************************************************"
         (match espace_discussion with
             | Espace_non_anonyme nom -> nom
             | Espace_anonyme nom -> nom ^ extension_anonyme
             | Espace_prive (nom, _) -> nom ^ extension_prive)
         date
         heure
         "************************************************************************" ;
      for i = 0 to panneau_messages#rows - 1 do
         match espace_discussion with
            | Espace_anonyme _ ->
                 chaine := !chaine ^ (Printf.sprintf "%s\n" (panneau_messages#cell_text i 1))
            | Espace_non_anonyme _
            | Espace_prive _ ->
                 chaine := !chaine ^ (Printf.sprintf "%s : %s\n"
                                         (panneau_messages#cell_text i 0)
                                         (panneau_messages#cell_text i 1))
      done ;
      chaine := !chaine ^ (Printf.sprintf
                              "\n\n%s fin de la discussion %s\n\n\n\n"
                              "*************************"
                              "*************************") ;
      !chaine


   initializer
      panneau_messages#set_column ~auto_resize:true 0 ;
      begin
         match espace_discussion with
            | Espace_anonyme _ ->
                 panneau_messages#set_column ~visibility:false 0
            | _ ->
                 ()
      end ;
      panneau_noms#set_column ~visibility:false 0


end ;;



class gerant_espaces =
object (self)

   (* la liste des espaces connects *)
   val mutable liste_espaces_connectes = ([] : (espace_discussion * espace) list)

   (* l'espace courant *)
   val mutable espace_courant = Espace_anonyme ""

   (* permet de rcuprer le nombre d'espaces connects *)
   method nombre_espaces_connectes = List.length liste_espaces_connectes

   (* permet de rcuprer la liste des espaces connects *)
   method liste_espaces_connectes = fst (List.split liste_espaces_connectes)

   (* permet de rcuprer un espace, de type espace_discussion *)
   method espace_numero = function indice ->
      if indice >= List.length liste_espaces_connectes then
         failwith "grant espaces : espace numro impossible"
      else
         fst (List.nth liste_espaces_connectes indice)

   (* permet de rcuprer l'espace courant *)
   method espace_courant = espace_courant

   (* permet de dfinir l'espace courant *)
   method definir_espace_courant = function espace ->
      if not (List.mem_assoc espace liste_espaces_connectes) then
         failwith "grant espaces : dfinition espace courant impossible"
      else
         espace_courant <- espace

   (* permet de rcuprer la position d'un espace dans la liste *)
   method position_espace = function espace ->
      if not (List.mem_assoc espace liste_espaces_connectes) then
         failwith "grant espaces : position espace impossible"
      else
         let liste_espaces = fst (List.split liste_espaces_connectes)
         in
         let rec chercher = function 
            | [] -> failwith "grant espaces : position espace impossible"
            | [e] -> 0
            | e :: reste ->
                 if espace = e then 0 else 1 + (chercher reste)
         in
         chercher liste_espaces

   (* permet de rcuprer la discussion d'un espace *)
   method discussion = function espace ->
      if not (List.mem_assoc espace liste_espaces_connectes) then
         failwith "grant espaces : dfinition espace courant impossible"
      else
         (List.assoc espace liste_espaces_connectes)#discussion

   (* permet d'ajouter un espace  la liste *)
   method ajouter_espace = function espace ->
      let nouvel_espace = new espace espace
      in
      if not (List.mem_assoc espace liste_espaces_connectes) then
         begin
            liste_espaces_connectes <- (espace, nouvel_espace) :: liste_espaces_connectes ;
            liste_espaces_connectes <-
               List.sort 
                  (fun (e1, _) -> fun (e2, _) ->
                      let nom_espace_e1 = self#nom_espace e1 
                      and nom_espace_e2 = self#nom_espace e2 
                      in
                      if e1 < e2 then -1 else if e1 = e2 then 0 else 1)
                  liste_espaces_connectes
         end

   (* permet de rcuprer le nom d'un espace *)
   method nom_espace = function espace ->
      match espace with
         | Espace_anonyme nom
         | Espace_non_anonyme nom
         | Espace_prive (nom, _) -> nom

   (* permet d'ajouter plusieurs espaces *)
   method ajouter_espaces = function espaces ->
      List.iter (fun e -> self# ajouter_espace e) espaces

   (* permet de supprimer un espace de la liste *)
   method supprimer_espace = function espace ->
      if List.mem_assoc espace liste_espaces_connectes then
         liste_espaces_connectes <- List.remove_assoc espace liste_espaces_connectes

   (* permet d'ajouter un client dans les espaces spcifis *)
   method ajouter_client = function client -> function espaces ->
      List.iter (fun e -> 
                    if List.mem_assoc e liste_espaces_connectes then
                       (List.assoc e liste_espaces_connectes)#ajouter_client client
                    else
                       failwith "grant espaces : ajout client impossible") espaces

   (* permet de supprimer un client dans les espaces spcifis *)
   method supprimer_client = function client -> function espaces ->
      List.iter (fun e -> 
                    if List.mem_assoc e liste_espaces_connectes then
                       (List.assoc e liste_espaces_connectes)#supprimer_client client
                    else
                       failwith "grant espaces : suppression client impossible") espaces


   (* permet de rcuprer le composant graphique d'affichage des messages d'un espace *)
   method panneau_messages = function espace ->
      if not (List.mem_assoc espace liste_espaces_connectes) then
         failwith "grant espaces : panneau messages impossible"
      else
         (List.assoc espace liste_espaces_connectes)#panneau_messages

   (* permet de rcuprer le composant graphique d'affichage des noms d'un espace *)
   method panneau_noms = function espace ->
      if not (List.mem_assoc espace liste_espaces_connectes) then
         failwith "grant espaces : panneau noms impossible"
      else
         (List.assoc espace liste_espaces_connectes)#panneau_noms


   (* permet d'attribuer un rflexe au panneau des messages d'un espace *)
   method attribuer_reflexe = function espace -> function reflexe ->
      if not (List.mem_assoc espace liste_espaces_connectes) then
         failwith "grant espaces : attribuer rflexe impossible"
      else
         (List.assoc espace liste_espaces_connectes)#attribuer_reflexe reflexe

   (* permet de rcuprer le nombre de lignes d'une discussion *)
   method nombre_lignes_discussion = function espace ->
      if not (List.mem_assoc espace liste_espaces_connectes) then
         failwith "grant espaces : nombre lignes discussion impossible"
      else
         (List.assoc espace liste_espaces_connectes)#nombre_lignes

   (* permet d'effacer la discussion d'un espace *)
   method effacer_discussion = function espace ->
      if not (List.mem_assoc espace liste_espaces_connectes) then
         failwith "grant espaces : effacer discussion impossible"
      else
         (List.assoc espace liste_espaces_connectes)#effacer_messages ()

   (* permet d'afficher un message sur l'espace spcifi *)
   method afficher_message = function espace -> function nom -> function message ->
      if not (List.mem_assoc espace liste_espaces_connectes) then
         failwith "grant espaces : afficher message impossible"
      else
         (List.assoc espace liste_espaces_connectes)#afficher_message nom message

   (* permet d'afficher un message sur tous les espaces connects *)
   method afficher_message_partout = function nom -> function message ->
      List.iter 
         (fun e -> self#afficher_message e nom message) 
         (fst (List.split liste_espaces_connectes))
      
   (* permet d'initialiser les espaces graphiques en spcifiant les espaces et les clients *)
   method initialiser_liste_espaces_graphiques = 
      let rec initialiser_liste = function
         | [] -> []
         | espace :: reste -> (espace, (new espace espace)) :: (initialiser_liste reste)
      in
      let rec attribuer_clients = function
         | [] -> ()
         | (client, espaces) :: reste ->
              List.iter
                 (fun e -> (List.assoc e liste_espaces_connectes)#ajouter_client client)
                 espaces ;
                 attribuer_clients reste
      in
      function espaces -> function liste_clients ->
         liste_espaces_connectes <- initialiser_liste espaces ;
         attribuer_clients liste_clients


end ;;
