(*#####################################################################*
 *
 * Module serveur pour le service de messagerie
 *
 * Toute la gestion du serveur y est inclue.
 *
 *#####################################################################*)


(***********************************************************************
 * Chargement des modules
 ***********************************************************************)
open Types ;;
open Lexeur ;;
open Parseur ;;
open Cryptage ;;


let rec imprimer_liste_espaces = function
	| [] -> flush stdout

	| Espace_anonyme nom :: reste ->
		  print_endline (nom ^ " (anonyme)") ;
		  imprimer_liste_espaces reste

	| Espace_non_anonyme nom :: reste ->
		  print_endline nom ;
		  imprimer_liste_espaces reste

	| Espace_prive (nom, _) :: reste ->
		  print_endline (nom ^ " (priv)") ;
		  imprimer_liste_espaces reste
;;



(***********************************************************************
 * Variables globales
 ***********************************************************************)

(* liste permettant de grer tous les clients connects au serveur.
 * aura le type ((int * client) list) ref *)
let liste_clients = ref [] ;;

let chemin_config =
   if Sys.os_type = "Unix" then
      (Sys.getenv "HOME") ^ "/.otchatche/"
   else
      ""
;;

(* le nom du fichier de configuration *)
let nom_fichier_config = chemin_config ^ "serveur.conf" ;;

(* liste grant les espaces de discussion. On peut considrer que "potins" est
 * un espace dj existant, par dfaut, et qu'on ne peut pas le supprimer. *)
let resultat_parsing =
   try 
      let fichier = open_in nom_fichier_config in
      let triplet = analyser_espaces lexeur (Lexing.from_channel fichier)
      in
      close_in fichier ;
      triplet
   with
      | Lexeme_inconnu (numero, lexeme) ->
           Printf.printf "%s : erreur  la ligne %d.\n"
              nom_fichier_config
              numero ;
           Printf.printf "%s : %s n'appartient pas au langage.\n\n"
              nom_fichier_config
              lexeme ;
           flush stdout ;
           exit 2

      | Fin_ligne_inattendue numero ->
           Printf.printf "%s : erreur  la ligne %d.\n"
              nom_fichier_config
              numero ;
           Printf.printf "%s : fin de ligne inattendue.\n\n"
              nom_fichier_config ;
           flush stdout ;
           exit 2
;;

let port_client, liste_espaces_permanents, liste_messages = resultat_parsing ;;

let liste_espaces_discussion = ref liste_espaces_permanents ;;


(* donnes relatives au service de messagerie *)
let maximum_connexions = 1000 ;;

(* mutex pour la protection des donnes paratges *)
let mutex_liste_clients = Mutex.create () ;;
let mutex_liste_espaces = Mutex.create () ;; 

(* variable ncessaire pour tuer le serveur lorsque l'on quitte le programme. *)
let processus_clients = ref (Thread.self ()) ;;  (* valeur relle *)


(***********************************************************************
 * Fonctions de service
 ***********************************************************************)

(* renvoie un numro de client unique *)
let numero_client =
   let numero_initial = 100 in
   let numero = ref numero_initial
   in
   function () ->
      incr numero ;
      if !numero = max_int then
         numero := numero_initial ;
      !numero
;;

(* calcule l'intersection de deux listes *)
let rec intersection_listes = function l1 -> function
   | [] -> []
   | element :: reste ->
        if List.mem element l1 then
           element :: (intersection_listes l1 reste)
        else
           intersection_listes l1 reste
;;

(* permet de calculer les clients connects  un espace *)
let calculer_client_espace =
   let rec calculer = function espace -> function
      | [] -> 0
      | (_, client) :: reste ->
           let espaces_client = client#liste_espaces
           in
           if List.mem espace espaces_client then
              begin
                 1 + (calculer espace reste)
              end
           else
              (calculer espace reste)
   in
   function espace ->
      calculer espace !liste_clients
;;

(* teste si un espace temporaire a encore des connexions *)
let supprimer_espace_temporaire = (*function client ->*) function espace ->

   Printf.printf "Je reois une dconnexion de l'espace %s.\n"
      (match espace with Espace_anonyme nom | Espace_non_anonyme nom | Espace_prive (nom, _) -> nom) ;
   flush stdout ;
   if not (List.mem espace liste_espaces_permanents) then
      begin
         Printf.printf "L'espace n'est pas un espace permanent.\n" ;
         flush stdout ;
         let nb_clients = calculer_client_espace espace
         in
         Printf.printf "Le nombre de clients connects est %d.\n" nb_clients ;
         flush stdout ;
         if nb_clients = 0 then
            begin
               Printf.printf "Suppression de l'espace.\n" ;
               flush stdout ;
               Mutex.lock mutex_liste_espaces ;
               liste_espaces_discussion :=
                  List.filter (fun e -> e <> espace) !liste_espaces_discussion ;
               Mutex.unlock mutex_liste_espaces ;
               Printf.printf "Diffusion de la suppression de l'espace.\n\n" ;
               flush stdout (*;
               client#distribuer (Notification_suppression_espace espace)*)
            end
      end
   else
      begin
         Printf.printf "L'espace est un espace permanent.\n\n" ;
         flush stdout
      end
;;

let supprimer_espaces_temporaires = function client ->
   List.iter
      (function espace ->
          if not (List.mem espace liste_espaces_permanents) then
             supprimer_espace_temporaire (*client*) espace)
      client#liste_espaces
;;

(***********************************************************************
 * Dfinition de types
 ***********************************************************************)
class client =
object (self)

   (*------------------- Champs de la classe client -------------------------*)
   val mutable numero = 0                  (* identifiant du client          *)
   val mutable nom = ""                    (* nom du client                  *)
   val mutable canal_entree = stdin        (* canal d'entre du client       *)
   val mutable canal_sortie = stdout       (* canal de sortie du client      *)
   val mutable processus = Thread.self ()  (* s'occupe de la lecture des v-
                                            * nements envoys par le client  *)
   val mutable liste_espaces = []          (* liste des espaces de discussion*)


   (*------------------ mthodes d'accs aux champs du client ---------------*)
   method numero   = numero
   method identite = Identite (numero, nom)
   method liste_espaces = liste_espaces
   method nom = nom

   method set_liste_espaces = function liste ->
      liste_espaces <- liste

   (*-------- mthode de cration et de destruction d'instance --------------*)
   method demarrer = function numero_client -> function ce -> function cs ->
      numero        <- numero_client ;
      canal_entree  <- ce ;
      canal_sortie  <- cs ;
      liste_espaces <- !liste_espaces_discussion ;
      processus     <- Thread.create self#lire_evenements_mode_init ()

   (* permet de tuer le client en fermant ses canaux de communication 
    * et en tuant le processus qui lui est associ. *)
   method tuer = function () ->
      close_in canal_entree ;
      close_out canal_sortie ;
      Mutex.lock mutex_liste_clients ;
      liste_clients := List.remove_assoc numero !liste_clients ;
      Mutex.unlock mutex_liste_clients ;
      supprimer_espaces_temporaires self ;
      Thread.kill processus

   (*--------------- mthodes de rception et d'envoi d'vnements ----------*)
   (*--------------------------- relatives  un client ----------------------*)
   method envoyer_evenement = function (evenement : evenement) ->
      ecrire_evenement canal_sortie evenement

   method private recevoir_evenement = function () ->
      lire_evenement canal_entree

   (* mthode permettant de construire la liste des clients  envoyer au client
    * courant *)
   method private liste_identites_et_espaces_communs =
      let rec boucle = function
         | [] -> []
         | (_, (client : client)) :: reste ->
              if client#nom = "" then
                 boucle reste
              else
                 (client#identite, intersection_listes client#liste_espaces liste_espaces)
                 :: (boucle reste)
      in
      function () ->
         boucle !liste_clients


   (* mthode permettant de construire la liste des clients  envoyer au client
    * courant *)
   method private liste_identites_et_espaces =
      let rec boucle = function
         | [] -> []
         | (_, (client : client)) :: reste ->
              if client#nom = "" then
                 boucle reste
              else
                 (client#identite, client#liste_espaces) :: (boucle reste)
      in
      function () ->
         boucle !liste_clients


   (* mthode permettant de couper la connexion avec le client *)
   method private couper_connexion = function message ->
      print_endline message ;
      flush stdout ;
      self#tuer ()


   (* distribue le message  tous les clients *)
   method distribuer =
      let rec boucle = function evenement -> function
         | [] -> ()         
         | (_, (client : client)) :: reste ->
              match evenement with
                 | Arrivee_client (id, liste)  ->
                      let intersection = intersection_listes liste client#liste_espaces
                      in
                      if intersection <> [] then 
                         client#envoyer_evenement (Arrivee_client (id, intersection)) ;
                      boucle evenement reste

                 | Depart_client (id, liste) ->
                      let intersection = intersection_listes liste client#liste_espaces
                      in
                      if intersection <> [] then
                         client#envoyer_evenement (Depart_client (id, intersection)) ;
                      boucle evenement reste

                 | Envoi_message (_, espace, _)  ->
                      if List.mem espace client#liste_espaces then
                         client#envoyer_evenement evenement ;
                      boucle evenement reste

                 | Changement_espace (_, espace_depart, espace_arrivee) ->
                      if List.mem espace_depart client#liste_espaces || 
                         List.mem espace_arrivee client#liste_espaces then
                            client#envoyer_evenement evenement ;
                      boucle evenement reste

                 | Notification_nouvel_espace _ ->
                      client#envoyer_evenement evenement ;
                      boucle evenement reste

                 | Notification_suppression_espace _ ->
                      client#envoyer_evenement evenement ;
                      boucle evenement reste

                 | Connexion_a (id, liste) ->
                      let intersection = intersection_listes liste client#liste_espaces
                      in
                      if intersection <> [] then
                         client#envoyer_evenement (Connexion_a (id, intersection)) ;
                      boucle evenement reste

                 | Deconnexion_de (_, espace) ->
                      if List.mem espace client#liste_espaces then
                         client#envoyer_evenement evenement ;
                      boucle evenement reste

                 | _ ->
                      print_endline ("serv_" ^ nom ^ ": je reois un vnement "
                                     ^ "qu'un client ne devrait pas pouvoir distribuer.") ;
                      flush stdout ;
                      failwith "cas impossible"
      in
      function evenement ->
         boucle evenement !liste_clients


   (* permet de lire les messages pendant le mode d'initialisation *)
   method private lire_evenements_mode_init = function () ->

      (* on envoie le numro du client et la liste des espaces disponibles *)
      print_endline "serv_??? : j'envoie le numro, la liste des espaces" ;
      print_endline "           et la liste des noms." ;
      flush stdout ;
      ecrire_evenement canal_sortie (Envoi_numero numero) ;
      ecrire_evenement canal_sortie (Envoi_liste_espaces !liste_espaces_discussion) ;
      ecrire_evenement canal_sortie (Envoi_liste_clients (self#liste_identites_et_espaces ())) ;

      Printf.printf "\nserv_??? : j'attends un vnement (mode init).\n" ;
      flush stdout ;
      try
         begin
            let evenement_1 = self#recevoir_evenement ()
            in
            match evenement_1 with
               | Depart_client _ ->
                    Printf.printf "serv_??? : dpart du client (mode init).\n" ; 
                    flush stdout ;
                    self#tuer () 

               | Envoi_nom nom_client ->
                    let evenement_espace = self#recevoir_evenement ()
                    in
                    Printf.printf "\nserv_??? : j'ai reu un vnement (mode init).\n" ;
                    flush stdout ;
                    begin
                       match evenement_espace with
                          | Envoi_liste_espaces nouvelle_liste ->
                               (* mises  jour des donnes importantes *)
                               nom <- nom_client ;
                               liste_espaces <- nouvelle_liste ;
                    
                               (* distribution de l'arrive du client dans le service *)
                               let evenement_arrivee = Arrivee_client (self#identite, nouvelle_liste)
                               in
                               Printf.printf "\nserv_%s : on distribue l'arrive  tous les clients." nom ;
                               flush stdout ;
                               self#distribuer evenement_arrivee ;
                               self#envoyer_messages_espaces () ;
                               self#lire_evenements ()

                          | _ ->
                               let message = Printf.sprintf 
                                                "serv_%s : vnement non attendu (mode init).\n"
                                                nom
                               in
                               self#couper_connexion message
                    end
                    
               | _ ->
                    let message = Printf.sprintf 
                                     "serv_%s : vnement non attendu (mode init).\n"
                                     nom
                    in
                    self#couper_connexion message
         end
      with
         | End_of_file ->
              self#tuer ()

   (* envoie tous les messages des espaces auxquels le client est connect *)
   method private envoyer_messages_espaces = function () ->
      List.iter
         (function (e, m) ->
             if List.mem e liste_espaces then
                let nom_espace = 
                   match e with
                      | Espace_anonyme nom
                      | Espace_non_anonyme nom
                      | Espace_prive (nom, _) -> nom
                in
                match List.assoc e liste_messages with
                   | None -> ()
                   | Some message ->
                        let message_crypte = crypter message
                        in
                        self#envoyer_evenement (Envoi_message (Identite (0, nom_espace), e, message_crypte)))
         liste_messages


   (* effectue la lecture des messages du client et les traite *)
   method private lire_evenements = function () ->
      try
         while true do
            Printf.printf "\nserv_%s : j'attends un vnement.\n" nom ;
            flush stdout ;
            let evenement = self#recevoir_evenement ()
            in
            Printf.printf "\nserv_%s : j'ai reu un vnement.\n" nom ;
            flush stdout ;

            (* on traite l'vnement*)
            match evenement with
               | Envoi_message (_, _, message) ->
                    Printf.printf "serv_%s : l'vnement est Envoi_message (%s).\n" nom message ;
                    flush stdout ;
                    (* on distribue l'information aux alentours *)
                    self#distribuer evenement

               | Changement_espace _ ->
                    Printf.printf "serv_%s : l'vnement est Changement_espace.\n" nom ;
                    flush stdout ;
                    self#distribuer evenement

               | Depart_client _ ->
                    Printf.printf "serv_%s : l'vnement est Depart_client.\n" nom ;
                    flush stdout ;
                    self#distribuer evenement ;
                    self#tuer ()

               | Demande_liste_clients ->
                    Printf.printf "serv_%s : l'vnement est Demande_liste_clients.\n" nom ;
                    flush stdout ;
                    self#envoyer_evenement
                       (Envoi_liste_clients (self#liste_identites_et_espaces_communs ()))

               | Demande_liste_espaces ->
                    Printf.printf "serv_%s : l'vnement est Demande_liste_espaces.\n" nom ;
                    flush stdout ;
                    self#envoyer_evenement
                       (Envoi_liste_espaces !liste_espaces_discussion)

               | Demande_nouvel_espace espace ->
                    Mutex.lock mutex_liste_espaces ;
                    if List.mem espace !liste_espaces_discussion then
                       begin
                          self#envoyer_evenement (Refus_creation_espace espace) ;
                          Mutex.unlock mutex_liste_espaces
                       end      
                    else
                       begin
                          liste_espaces_discussion := !liste_espaces_discussion @ [espace] ;
                          Mutex.unlock mutex_liste_espaces ;
                          self#envoyer_evenement (Acceptation_creation_espace espace) ;
                          self#distribuer (Notification_nouvel_espace espace)
                       end

               | Envoi_liste_espaces nouvelle_liste ->
                    liste_espaces <- nouvelle_liste

               | Connexion_a _ ->
                    self#distribuer evenement

               | Deconnexion_de (_, espace) ->
                    supprimer_espace_temporaire espace ;
                    self#distribuer evenement

               | _ ->
                    Printf.printf "serv_%s : vnement impossible pour un client.\n" nom ;
                    flush stdout ;
                    failwith (Printf.sprintf "serv_%s : cas impossible" nom) ;
         done
      with
         | End_of_file | Sys_error _ ->
            (* le client a coup la connexion de faon brutale, on le
               retire de la liste. *)
            Printf.printf "serv_%s : le client est parti sans prvenir.\n" nom ;
            flush stdout ;
            let evenement = Depart_client (Identite (numero, nom), liste_espaces)
            in
            Mutex.lock mutex_liste_clients ;
            liste_clients := List.remove_assoc numero !liste_clients ;
            Mutex.unlock mutex_liste_clients ;
            self#distribuer evenement ;
            self#tuer ()

         | exn ->
            Printf.printf "serv_%s : exception inattendue.\n" nom ;
            print_endline ("           " ^ (Printexc.to_string exn)) ;
            flush stdout

end ;;


(***********************************************************************
 * Fonctions de la messagerie
 ***********************************************************************)

(* lance l'coute de connexions pour les clients *)
let rec serveur_client = function sock ->
   let in_channel = ref stdin (*  l'extrieur pour pouvoir clore la connexion
                               * du client. *)
   in
   try
      while true do

         print_endline "\nserv_clients : je me mets en attente de clients." ;
         flush stdout ;

         (* on rcupre les canaux du client *)
         let sock_client = fst (ThreadUnix.accept sock)
         in

         print_endline "\nserv_clients : je viens de recevoir une connexion." ;
         flush stdout ;

         (* on remplit les champs du nouveau client *)
         in_channel := Unix.in_channel_of_descr sock_client ;
         let out_channel = Unix.out_channel_of_descr sock_client
         and numero = numero_client () in
         let client = new client
         in

         (* ajout du client dans la liste *)
         print_endline "\nserv_clients : on ajoute le client  la liste." ;
         flush stdout ;
         Mutex.lock mutex_liste_clients ;
         liste_clients := (numero, client) :: !liste_clients ;
         Mutex.unlock mutex_liste_clients ;

         print_endline "serv_clients : je lance le nouveau client." ;
         flush stdout ;
         client#demarrer numero !in_channel out_channel ;

      done

   with
      | Unix.Unix_error (erreur, chaine, arguments) ->
           print_endline "serveur : une erreur systme (unix/windows) s'est produite..." ;
           print_endline
              ("serveur : " ^ (Printexc.to_string (Unix.Unix_error (erreur, chaine, arguments)))) ;
           print_endline ("serveur : " ^ (Unix.error_message erreur)) ;
           flush stdout ;
           Unix.shutdown_connection !in_channel ;
           serveur_client sock
      | exn ->
           print_endline "serveur : une erreur inattendue s'est produite..." ;
           print_endline ("serveur : " ^ (Printexc.to_string exn)) ;
           flush stdout ;
           Unix.shutdown_connection !in_channel ;
           serveur_client sock
;;

(***********************************************************************
 * Fonction de lancement du serveur
 ***********************************************************************)

(* lance le serveur des clients : elle rcupre les informations
 * locales, attache le serveur  l'adresse locale, puis se met en
 * attente des clients. *)
let lancer_serveur = function _ ->
   try

      (* le serveur de messagerie affiche les informations locales *)
      print_endline ("\nserveur : voici les informations locales...") ;
      print_endline ("\n\tport du serveur des clients : "
                     ^ (string_of_int port_client)) ;
      print_endline ("\n\tadresse ip : "
                     ^ (Unix.string_of_inet_addr (adresse_locale ())) ^ "\n") ;
      
      (* on doit tout d'abord crer la socket du serveur *)
      print_endline ("\nserveur : je cre les sockets du serveur des clients ") ;
      flush stdout ;
      
      let sock_client = ThreadUnix.socket Unix.PF_INET Unix.SOCK_STREAM 0
      in
      
      print_endline "serveur : j'attache le serveur." ;
      flush stdout ;
      
      (* on doit ensuite attacher le serveur  son adresse *)
      Unix.bind sock_client (Unix.ADDR_INET (adresse_locale (), port_client)) ;
      print_endline "serveur : j'attache le serveur de clients " ;
      flush stdout ;
      
      print_endline ("serveur : je spcifie le nombre de connexions simultannes "
                     ^ "par serveur.") ;
      flush stdout ;
      
      (* on doit ensuite spcifier le nombre de connexions que le serveur accepte *)
      Unix.listen sock_client maximum_connexions ;
      
      print_endline ("\nserveur : je continue en lanant la boucle d'coute des "
                     ^ "connexions des clients.") ;
      flush stdout ;
      
      (* on peut maintenant tourner "en dur" pour attendre les connexions des clients *)
      serveur_client sock_client
         
   with
      | Unix.Unix_error (erreur, "bind", _) ->
           print_endline "serveur : l'adresse du serveur est dj occupe!" ;
           print_endline ("serveur : " ^ (Unix.error_message erreur)) ;
           flush stdout ;
           exit 2
      | exn ->
           print_endline "serveur : une erreur inattendue s'est produite..." ;
           print_endline ("serveur : " ^ (Printexc.to_string exn)) ;
           flush stdout ;
           exit 2 
;;



lancer_serveur () ;;
