(*#####################################################################*
 *
 * Module de types pour le service de messagerie
 *
 * On y trouve tous les types et les fonctions basiques de communication
 * (envoi et rception d'vnements).
 *
 *#####################################################################*)


(***********************************************************************
 * Les types
 ***********************************************************************)

(* les informations concernant l'identit d'un client
 * Identit (numero_client, nom_client) *)
type identite_client = Identite of int * string
;;


(* les espaces de discussion sont dtermins par leurs noms. *)
type espace_discussion = 
   | Espace_anonyme of string
   | Espace_non_anonyme of string
   | Espace_prive of string * (string list)
;;


(* l'vnement est l'lment unitaire du protocole de communication. Ces
 * vnements sont transmis entre le client et le serveur de messagerie,
 * sachant tout de mme que soit l'un, soit l'autre, ne peut envoyer que
 * certains vnements. *)
type evenement =
   (* vnements uniquement envoys par le client :
    *    o Demande_liste_clients permet au client d'obtenir la liste des
    *      clients connects.
    *    o Demande_liste_espaces permet au client d'obtenir la liste des
    *      espaces de discussion ouverts.
    *    o Envoi_nom est envoy au serveur de messagerie pour 
    *      indiquer l'identit du client et ainsi recevoir un numro
    *      de client unique. 
    *    o Demande_nouvel_espace permet au client de demander au serveur
    *      de crer un nouvel espace de discussion. La demande peut tre
    *      refuse s'il existe un espace du mme nom.
    *    o Connexion_a permet d'indiquer au serveur que l'on vient de se
    *      connecter aux espaces dsigns.
    *    o Deconnexion_de permet d'indiquer les espaces desquels le client
    *      s'est dconnect. *)
   | Demande_liste_clients
   | Demande_liste_espaces
   | Envoi_nom of string
   | Demande_nouvel_espace of espace_discussion
   | Connexion_a of identite_client * (espace_discussion list)
   | Deconnexion_de of identite_client * espace_discussion

   (* vnements uniquement envoys par le serveur de messagerie. 
    *    o Envoi_numero est permet au serveur de communiquer le numro qui a t
    *      attribu au client qui vient de se connecter.
    *    o Arrivee_client permet au serveur de signaler l'arrive d'un
    *      nouveau client sur le service. On signale l'identit et les espaces
    *      o le client se trouve (pour pouvoir rpartir les arrives dans les
    *      espaces respectifs).
    *    o Envoi_liste_clients permet au serveur de communiquer la liste des
    *      clients et les espaces de discussion auxquels ils sont connects.
    *    o Notification_nouvel_espace permet au serveur d'indiquer  tous les
    *      clients qu'un nouvel espace de discussion est disponible.
    *    o Notification_suppression_espace permet au serveur d'indiquer 
    *      tous les clients qu'un espace de discussion vient d'tre ferm.
    *    o Refus_creation_espace permet au serveur de prvenir le demandeur
    *      du refus de sa demande de cration d'espace.
    *    o Refus_suppression_espace permet au serveur de prvenir le demandeur
    *      du refus de sa demande de suppression d'espace. *)
   | Envoi_numero of int
   | Arrivee_client of identite_client * (espace_discussion list)
   | Envoi_liste_clients of (identite_client * (espace_discussion list)) list
   | Notification_nouvel_espace of espace_discussion
   | Notification_suppression_espace of espace_discussion
   | Refus_creation_espace of espace_discussion
   | Acceptation_creation_espace of espace_discussion

   (* vnements envoys par les deux acteurs. 
    *    o Envoi_message envoie un message saisi par le client, et l'envoie avec        
    *      son identit et l'espace dans lequel il l'a crit. Le serveur utilise
    *      cet vnement pour retransmettre le message  tous ceux qui sont connects
    *      sur cet espace.
    *    o Depart_client est envoy par un client et transmis  tous les autres
    *      personnes connects aux mmes espaces pour signifier que l'auteur de
    *    o l'vnement est parti.
    *    o Envoi_liste_espaces permet au serveur d'envoyer la liste des espaces
    *      de discussion disponibles, et au client de transmettre ses choix au
    *      serveur.
    *    o Envoi_liste_noms permet au serveur d'envoyer la liste des noms dj
    *      pris (le client ne peut choisir un nom dj pris).
    *    o Changement_espace permet au client de transmettre aux autres clients
    *      (via le serveur) qu'il change temporairement de groupe de discussion. *)
   | Envoi_message of identite_client * espace_discussion * string
   | Depart_client of identite_client * (espace_discussion list)
   | Envoi_liste_espaces of espace_discussion list
   | Changement_espace of identite_client * espace_discussion * espace_discussion

;;


(* type dcrivant le bon droulement ou non du processus de lancement de
   connexion *)
type etat_connexion =
   | Correct
   | Erreur_serveur_messagerie
   | Erreur_serveur_intermediaire
   | Erreur_inattendue
;;


(***********************************************************************
 * Les fonctions de base en communication
 ***********************************************************************)

(* lire_evenement lit, sur un canal pass en argument, une valeur de type
 * evenement et la retourne en rsultat *)
let lire_evenement = function in_channel ->
   (input_value in_channel : evenement)
;;

(* ecrire_evenement envoie sur le canal pass en argument une valeur de type
 * vnement. Une fois avoir mis cette valeur sur le canal, on force l'envoi
 * par flush. *)
let ecrire_evenement = function out_channel -> function (evenement : evenement) ->
   output_value out_channel evenement ;
   flush out_channel
;;

(* adresse_locale retourne l'adresse internet correspondant  la machine
 * locale *)
let adresse_locale = function () ->
   try
      (Unix.gethostbyname (Unix.gethostname ())).Unix.h_addr_list.(0)
   with
      | Not_found -> Unix.inet_addr_of_string "127.0.0.1"
;;
